import json
from typing import Dict

from importlib_resources import as_file, files

from .smiles_language import SELFIESLanguage, SMILESLanguage, SMILESTokenizer  # noqa

with as_file(
    files('pytoda.smiles.metadata').joinpath('tokenizer', 'vocab.json')
) as vocab_filepath:
    with open(vocab_filepath, 'r') as f:
        vocab: Dict[str, int] = json.load(f)



"""Testing SetMatchingDataset."""
import unittest
from typing import List

import torch
from scipy.optimize import linear_sum_assignment
from torch.utils.data import DataLoader

from pytoda.datasets import (
    DistributionalDataset,
    PairedSetMatchingDataset,
    PermutedSetMatchingDataset,
)
from pytoda.datasets.utils.factories import (
    DISTRIBUTION_FUNCTION_FACTORY,
    METRIC_FUNCTION_FACTORY,
)

seeds = [None, 42]
distribution_seeds = [None, 42]

min_set_length = [5, 2]
set_padding_value = 6.0
dataset_size = 250
max_set_length = 10
item_shape = (max_set_length, 4)
cost_metric = 'p-norm'
cost_metric_args = {'p': 2}
distribution_type = ['normal', 'uniform']
distribution_args = [{'loc': 0.0, 'scale': 1.0}, {'low': 0, 'high': 1}]
noise_std = [0.001, 0.1]  # Pytorch >1.7 errors with a noise of 0.0.

cost_metric_function = METRIC_FUNCTION_FACTORY[cost_metric](**cost_metric_args)

permute = [True, False]
DATASET_FACTORY = {True: PermutedSetMatchingDataset, False: PairedSetMatchingDataset}


class TestSetMatchingDataset(unittest.TestCase):
    """Test SetMatchingDataset class."""

    def test_permuted_set_matching_dataset(self) -> None:
        """Test PermutedSetMatchingDataset class."""

        def tolist(x: torch.Tensor) -> List:
            return x.flatten().tolist()

        for dist_type, dist_args in zip(distribution_type, distribution_args):

            distribution_function = DISTRIBUTION_FUNCTION_FACTORY[dist_type](
                **dist_args
            )
            for dist_seed in distribution_seeds:
                for seed in seeds:
                    for noise in noise_std:
                        for min_len in min_set_length:

                            s1 = DistributionalDataset(
                                dataset_size,
                                item_shape,
                                distribution_function,
                                seed=dist_seed,
                            )
                            datasets = [s1]

                            permuted_dataset = PermutedSetMatchingDataset(
                                *datasets,
                                min_len,
                                cost_metric_function,
                                set_padding_value=set_padding_value,
                                noise_std=noise,
                                seed=seed,
                            )
                            # Test length
                            self.assertEqual(len(permuted_dataset), dataset_size)

                            # Test __getitem__
                            sample1 = permuted_dataset[0]
                            sample2 = permuted_dataset[0]

                            self.assertIsInstance(sample1, tuple)
                            self.assertEqual(len(sample1), 5)

                            if dist_seed is not None and seed is not None:
                                # since both distribution seed and permutation seed
                                # are fixed, sampling twice with same index should return
                                # identical samples.

                                for item1, item2 in zip(sample1, sample2):
                                    self.assertTrue(torch.equal(item1, item2))

                                self.assertTrue(
                                    torch.equal(
                                        sample1[1][sample1[2].long(), :],
                                        sample2[1][sample2[2].long(), :],
                                    )
                                )

                                # When noise =0, permutation of set2 should return set1
                                # and vice versa
                                if noise < 0.01:

                                    for a, b in zip(
                                        tolist(sample1[1][sample1[2].long(), :]),
                                        tolist(sample1[0]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)
                                    for a, b in zip(
                                        tolist(sample1[0][sample1[3].long(), :]),
                                        tolist(sample1[1]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                else:

                                    self.assertFalse(
                                        torch.equal(
                                            sample1[1][sample1[2].long(), :],
                                            sample1[0],
                                        )
                                    )
                                    self.assertFalse(
                                        torch.equal(
                                            sample1[0][sample1[3].long(), :],
                                            sample1[1],
                                        )
                                    )

                            elif dist_seed is not None:
                                # Since only distribution seed is fixed, for fixed length
                                # settings the reference set returned at index 0 must be identical.
                                # NOTE: since items are padded when lengths vary,
                                # a lower limit on max set length is required to test
                                # that permutations are not equal when permutation seed is None.

                                if min_len == max_set_length:
                                    self.assertTrue(torch.equal(sample1[0], sample2[0]))
                                    if max_set_length > 3:
                                        self.assertFalse(
                                            torch.equal(sample1[2], sample2[2]),
                                            msg=f'{sample1},{sample2}',
                                        )

                                    if noise < 0.01:
                                        for a, b in zip(
                                            tolist(sample1[1][sample1[2].long(), :]),
                                            tolist(sample2[1][sample2[2].long(), :]),
                                        ):
                                            self.assertAlmostEqual(a, b, places=2)

                                elif sample1[-1] != sample2[-1]:
                                    # reason for asserting false is that length
                                    # cropping is a random event dependent on
                                    # permutation seed which is None in this setting

                                    self.assertFalse(
                                        torch.equal(sample1[0], sample2[0]),
                                        msg=f'{sample1},{sample2}',
                                    )

                                self.assertFalse(
                                    torch.equal(sample1[1], sample2[1]),
                                )

                                if noise < 0.01:
                                    for a, b in zip(
                                        tolist(sample1[1][sample1[2].long(), :]),
                                        tolist(sample1[0]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                    for a, b in zip(
                                        tolist(sample1[0][sample1[3].long(), :]),
                                        tolist(sample1[1]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                else:
                                    self.assertFalse(
                                        torch.equal(
                                            sample1[1][sample1[2].long(), :], sample1[0]
                                        )
                                    )
                                    self.assertFalse(
                                        torch.equal(
                                            sample1[0][sample1[3].long(), :], sample1[1]
                                        )
                                    )

                            elif seed is not None:
                                # since distribution seed is None, the sampled sets
                                # must be different but the lengths and permutations
                                # should be the same since the permutation seed is set.

                                self.assertFalse(torch.equal(sample1[0], sample2[0]))
                                self.assertFalse(torch.equal(sample1[1], sample2[1]))

                                self.assertTrue(torch.equal(sample1[-1], sample2[-1]))
                                if noise < 0.01:

                                    for a, b in zip(
                                        tolist(sample1[2]),
                                        tolist(sample2[2]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                    for a, b in zip(
                                        tolist(sample1[1][sample1[2].long(), :]),
                                        tolist(sample1[0]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                    for a, b in zip(
                                        tolist(sample1[0][sample1[3].long(), :]),
                                        tolist(sample1[1]),
                                    ):
                                        self.assertAlmostEqual(a, b, places=2)

                                else:
                                    self.assertFalse(
                                        torch.equal(
                                            sample1[1][sample1[2].long(), :],
                                            sample1[0],
                                        )
                                    )
                                    self.assertFalse(
                                        torch.equal(
                                            sample1[0][sample1[3].long(), :],
                                            sample1[1],
                                        )
                                    )

                            else:
                                # since both seeds are none, the sets and cropped lengths
                                # must be different. Difference in permutations are only
                                # checked if length>3 due to item-wise padding.
                                self.assertFalse(torch.equal(sample1[0], sample2[0]))
                                self.assertFalse(torch.equal(sample1[1], sample2[1]))

    def test_paired_set_matching_dataset(self) -> None:
        """Test PairedSetMatchingDataset class."""
        # Similar reasoning with respect to seeds follows from above.
        # Main difference is that the hungarian assignments, i.e, the targets
        # returned are not tested since these assignments are not dependent on
        # the permutation seed but only on the pair of sets generated.

        noise = 0.0

        for dist_type, dist_args in zip(distribution_type, distribution_args):

            distribution_function = DISTRIBUTION_FUNCTION_FACTORY[dist_type](
                **dist_args
            )
            for dist_seed in distribution_seeds:
                for seed in seeds:
                    for min_len in min_set_length:

                        if dist_seed is not None:
                            seed_s1 = dist_seed
                            seed_s2 = dist_seed + 1
                        else:
                            seed_s1 = seed_s2 = dist_seed

                        s1 = DistributionalDataset(
                            dataset_size,
                            item_shape,
                            distribution_function,
                            seed=seed_s1,
                        )
                        datasets = [s1]

                        s2 = DistributionalDataset(
                            dataset_size,
                            item_shape,
                            distribution_function,
                            seed=seed_s2,
                        )
                        datasets.append(s2)

                        paired_dataset = PairedSetMatchingDataset(
                            *datasets,
                            min_len,
                            cost_metric_function,
                            set_padding_value=set_padding_value,
                            noise_std=noise,
                            seed=seed,
                        )
                        # Test length
                        self.assertEqual(len(paired_dataset), dataset_size)

                        # Test __getitem__
                        sample1 = paired_dataset[0]
                        sample2 = paired_dataset[0]

                        sample1_hungarian12 = linear_sum_assignment(
                            torch.cdist(sample1[0], sample1[1]).numpy()
                        )[1]

                        sample1_hungarian21 = linear_sum_assignment(
                            torch.cdist(sample1[1], sample1[0]).numpy()
                        )[1]

                        self.assertTrue(
                            torch.equal(
                                sample1[2].int(),
                                torch.from_numpy(sample1_hungarian12).int(),
                            )
                        )
                        self.assertTrue(
                            torch.equal(
                                sample1[3].int(),
                                torch.from_numpy(sample1_hungarian21).int(),
                            )
                        )

                        self.assertIsInstance(sample1, tuple)
                        self.assertEqual(len(sample1), 5)
                        self.assertFalse(torch.equal(sample1[0], sample1[1]))

                        if dist_seed is not None and seed is not None:

                            for item1, item2 in zip(sample1, sample2):
                                self.assertTrue(torch.equal(item1, item2))

                        elif dist_seed is not None:

                            if min_len == max_set_length:
                                for item1, item2 in zip(sample1, sample2):
                                    self.assertTrue(torch.equal(item1, item2))

                            elif sample1[-1] != sample2[-1]:
                                for item1, item2 in zip(sample1[:2], sample2[:2]):
                                    self.assertFalse(torch.equal(item1, item2))

                        elif seed is not None:

                            self.assertFalse(torch.equal(sample1[0], sample2[0]))
                            self.assertFalse(torch.equal(sample1[1], sample2[1]))

                            self.assertTrue(torch.equal(sample1[-1], sample2[-1]))

                        else:
                            for item1, item2 in zip(sample1[:2], sample2[:2]):
                                self.assertFalse(torch.equal(item1, item2))

    def test_data_loader(self) -> None:
        """Test data_loader or SetMatchingDataset."""

        for dist_seed in distribution_seeds:
            for dist_type, dist_args in zip(distribution_type, distribution_args):
                distribution_function = DISTRIBUTION_FUNCTION_FACTORY[dist_type](
                    **dist_args
                )
                for noise in noise_std:
                    for seed in seeds:
                        for permute_ in permute:
                            for min_len in min_set_length:
                                if dist_seed is None:
                                    seed_s1 = seed_s2 = dist_seed

                                else:
                                    seed_s1 = dist_seed
                                    seed_s2 = dist_seed + 1

                                s1 = DistributionalDataset(
                                    dataset_size,
                                    item_shape,
                                    distribution_function,
                                    seed=seed_s1,
                                )
                                datasets = [s1]
                                if not permute_:
                                    s2 = DistributionalDataset(
                                        dataset_size,
                                        item_shape,
                                        distribution_function,
                                        seed=seed_s2,
                                    )
                                    datasets.append(s2)

                                setmatch_dataset = DATASET_FACTORY[permute_](
                                    *datasets,
                                    min_len,
                                    cost_metric_function,
                                    set_padding_value=set_padding_value,
                                    noise_std=noise,
                                    seed=seed,
                                )

                                data_loader = DataLoader(
                                    setmatch_dataset,
                                    batch_size=25,
                                )

                                for batch_index, batch in enumerate(data_loader):
                                    (
                                        set1_batch,
                                        set2_batch,
                                        idx12_batch,
                                        idx21_batch,
                                        len_batch,
                                    ) = batch

                                    self.assertEqual(
                                        set1_batch.shape,
                                        (
                                            25,
                                            max_set_length,
                                            4,
                                        ),
                                    )
                                    self.assertTrue(
                                        torch.unique(set1_batch, dim=0).size(0) == 25
                                    )
                                    self.assertTrue(
                                        torch.unique(set2_batch, dim=0).size(0) == 25
                                    )

                                    self.assertEqual(set1_batch.shape, set2_batch.shape)

                                    self.assertEqual(
                                        idx12_batch.shape, idx21_batch.shape
                                    )
                                    self.assertEqual(len(len_batch), 25)

                                    self.assertFalse(
                                        torch.equal(set1_batch, set2_batch)
                                    )

                                    if permute_ and noise_std == 0.0:

                                        ordered_set1 = set1_batch[
                                            torch.arange(0, idx21_batch.size(0))
                                            .unsqueeze(1)
                                            .repeat((1, idx21_batch.size(1))),
                                            idx21_batch,
                                        ]

                                        ordered_set2 = set2_batch[
                                            torch.arange(0, idx12_batch.size(0))
                                            .unsqueeze(1)
                                            .repeat((1, idx12_batch.size(1))),
                                            idx12_batch,
                                        ]

                                        self.assertTrue(
                                            torch.equal(set1_batch, ordered_set2)
                                        )
                                        self.assertTrue(
                                            torch.equal(ordered_set1, set2_batch)
                                        )


if __name__ == '__main__':
    unittest.main()



"""Testing ProteinFeatureLanguage."""
import unittest

from pytoda.proteins.processing import (
    AA_FEAT,
    AA_PROPERTIES_NUM,
    BLOSUM62,
    BLOSUM62_NORM,
    HUMAN_KINASE_ALIGNMENT_VOCAB,
    IUPAC_CODES,
)


class TestProcessing(unittest.TestCase):
    """Testing Processing."""

    def test__key_availability(self) -> None:
        """Test that the same keys are available in all alphabets."""
        self.assertEqual(AA_PROPERTIES_NUM.keys(), AA_FEAT.keys())
        self.assertEqual(BLOSUM62_NORM.keys(), BLOSUM62.keys())
        self.assertEqual(AA_PROPERTIES_NUM.keys(), BLOSUM62.keys())
        iupac = [x for x in IUPAC_CODES.values()] + [
            '<PAD>',
            '<START>',
            '<STOP>',
            '<UNK>',
            '-',
        ]
        self.assertEqual(iupac, list(BLOSUM62.keys()))
        self.assertTrue('-' in HUMAN_KINASE_ALIGNMENT_VOCAB)
        self.assertTrue('-' in BLOSUM62_NORM)
        self.assertTrue('-' in AA_PROPERTIES_NUM)

    def test__length_equal(self) -> None:
        for i in range(len(list(AA_PROPERTIES_NUM.values()))):
            self.assertEqual(len(list(AA_PROPERTIES_NUM.values())[i]), 6)
            self.assertEqual(len(list(AA_FEAT.values())[i]), 8)
            self.assertEqual(len(list(BLOSUM62.values())[i]), 26)
            self.assertEqual(len(list(BLOSUM62_NORM.values())[i]), 22)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import torch

from pytoda.types import Any, Tensor, Tuple


def range_tensor(
    value_range: Any, repetitions: Tuple[int], device: torch.device
) -> Tensor:
    """Returns a background tensor filled with a given range of values.

    Args:
        value_range (Any): Range of values to insert into each row of the
            background tensor.
        repetitions (Tuple[int]): The number of repetitions of value_range along each axis.
        device (torch.device): Device where the tensors are stored.

    Returns:
        Tensor: Tensor containing repetitions of the given range of values along specified axes.
            Example, value_range = [0,1,2], repetitions = (1,2) will repeat [0,1,2]
            once along dim 0 and twice along dim 1, i.e, tensor([[0,1,2,0,1,2]])
            of size (1,6) will be the output.
        NOTE: if a pattern [0,1,2] is required to fill a tensor of shape (2,5)
              specify value_range as [0,1,2,0,1] and repetiitons as (2,). The
              value_range is filled 'row-wise'. Simply transpose the output for
              a 'column-wise' fill.
    """
    return torch.from_numpy(np.tile(value_range, repetitions)).to(device)


def constant_value_tensor(value: float, shape: Tuple, device: torch.device) -> Tensor:
    """Returns a background tensor filled with a constant value.

    Args:
        value (float): Value to fill the background tensor with.
        shape (Tuple): Shape of the background tensor.
        device (torch.device): Device where the tensors are stored.

    Returns:
        Tensor: Tensor of given shape filled with the given constant value.
    """
    return torch.full(shape, value, device=device)



import torch

from pytoda.datasets.utils.background_tensors import constant_value_tensor, range_tensor
from pytoda.datasets.utils.wrappers import WrapperCDist, WrapperKLDiv

METRIC_FUNCTION_FACTORY = {'p-norm': WrapperCDist, 'KL': WrapperKLDiv}

DISTRIBUTION_FUNCTION_FACTORY = {
    'normal': torch.distributions.normal.Normal,
    'beta': torch.distributions.beta.Beta,
    'uniform': torch.distributions.uniform.Uniform,
    'bernoulli': torch.distributions.bernoulli.Bernoulli,
}

BACKGROUND_TENSOR_FACTORY = {
    'constant': constant_value_tensor,
    'range': range_tensor,
}



"""Implementation of DrugAffinityDataset."""
from typing import Iterable, Tuple

import pandas as pd
import torch
from torch.utils.data import Dataset

from ..proteins.protein_language import ProteinLanguage
from ..smiles.smiles_language import SMILESLanguage
from ..types import DrugAffinityData
from .protein_sequence_dataset import ProteinSequenceDataset
from .smiles_dataset import SMILESTokenizerDataset


class DrugAffinityDataset(Dataset):
    """
    Drug affinity dataset implementation.
    """

    def __init__(
        self,
        drug_affinity_filepath: str,
        smi_filepath: str,
        protein_filepath: str,
        column_names: Tuple[str] = ['ligand_name', 'sequence_id', 'label'],
        drug_affinity_dtype: torch.dtype = torch.int,
        smiles_language: SMILESLanguage = None,
        smiles_padding: bool = True,
        smiles_padding_length: int = None,
        smiles_add_start_and_stop: bool = False,
        smiles_augment: bool = False,
        smiles_canonical: bool = False,
        smiles_kekulize: bool = False,
        smiles_all_bonds_explicit: bool = False,
        smiles_all_hs_explicit: bool = False,
        smiles_randomize: bool = False,
        smiles_remove_bonddir: bool = False,
        smiles_remove_chirality: bool = False,
        smiles_vocab_file: str = None,
        smiles_selfies: bool = False,
        smiles_sanitize: bool = True,
        protein_language: ProteinLanguage = None,
        protein_amino_acid_dict: str = 'iupac',
        protein_padding: bool = True,
        protein_padding_length: int = None,
        protein_add_start_and_stop: bool = False,
        protein_augment_by_revert: bool = False,
        protein_randomize: bool = False,
        device: torch.device = (
            torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        ),
        iterate_dataset: bool = True,
        backend: str = 'eager',
    ) -> None:
        """
        Initialize a drug affinity dataset.

        Args:
            drug_affinity_filepath (str): path to drug affinity .csv file. Currently,
                the only supported format is .csv, with an index and three header
                columns named as specified in column_names.
            smi_filepath (str): path to .smi file.
            protein_filepath (str): path to .smi or .fasta file.
            column_names (Tuple[str]): Names of columns in data files to retrieve
                labels, ligands and protein name respectively.
                Defaults to ['ligand_name', 'sequence_id', 'label'].
            drug_affinity_dtype (torch.dtype): drug affinity data type.
                Defaults to torch.int.
            smiles_language (SMILESLanguage): a smiles language.
                Defaults to None.
            smiles_vocab_file (str): Optional .json to load vocabulary. Tries
                to load metadata if `iterate_dataset` is False.
                Defaults to None.
            smiles_padding (bool): pad sequences to longest in the smiles
                language. Defaults to True.
            smiles_padding_length (int): manually sets number of applied
                paddings, applies only if padding is True. Defaults to None.
            smiles_add_start_and_stop (bool): add start and stop token indexes.
                Defaults to False.
            smiles_canonical (bool): performs canonicalization of SMILES (one
                original string for one molecule), if True, then other
                transformations (augment etc, see below) do not apply
            smiles_augment (bool): perform SMILES augmentation. Defaults to
                False.
            smiles_kekulize (bool): kekulizes SMILES (implicit aromaticity
                only). Defaults to False.
            smiles_all_bonds_explicit (bool): Makes all bonds explicit.
                Defaults to False, only applies if kekulize = True.
            smiles_all_hs_explicit (bool): Makes all hydrogens explicit.
                Defaults to False, only applies if kekulize = True.
            smiles_randomize (bool): perform a true randomization of SMILES
                tokens. Defaults to False.
            smiles_remove_bonddir (bool): Remove directional info of bonds.
                Defaults to False.
            smiles_remove_chirality (bool): Remove chirality information.
                Defaults to False.
            smiles_selfies (bool): Whether selfies is used instead of smiles.
                Default to False.
            smiles_sanitize (bool): RDKit sanitization of the molecule.
                Defaults to True.
            protein_language (ProteinLanguage): protein language.
                Defaults to None, a.k.a., build it from scratch.
            protein_amino_acid_dict (str): Amino acid dictionary.
                Defaults to 'iupac'.
            protein_padding (bool): pad sequences to the longest in the protein
                language. Defaults to True.
            protein_padding_length (int): manually set the padding.
                Defaults to None.
            protein_add_start_and_stop (bool): add start and stop token
                indexes. Defaults to False.
            protein_augment_by_revert (bool): augment data by reverting the
                sequence. Defaults to False.
            protein_randomize (bool): perform a randomization of the protein
                sequence tokens. Defaults to False.
            device (torch.device): device where the tensors are stored.
                Defaults to gpu, if available.
            protein_vocab_file (str): Optional .json to load vocabulary. Tries
                to load metadata if `iterate_dataset` is False.
                Defaults to None.
            iterate_dataset (bool): whether to go through all items in the
                dataset to extend/build vocab, find longest sequence, and
                checks the passed padding length if applicable. Defaults to
                True.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
                Note that at the moment only the smiles dataset implement both
                backends. The drug affinity data and the protein dataset are
                loaded in memory.
        """
        Dataset.__init__(self)
        self.drug_affinity_filepath = drug_affinity_filepath
        self.smi_filepath = smi_filepath
        self.protein_filepath = protein_filepath
        # device
        self.device = device
        # backend
        self.backend = backend

        if not isinstance(column_names, Iterable):
            raise TypeError(f'Column names was {type(column_names)}, not Iterable.')
        if not len(column_names) == 3:
            raise ValueError(f'Please pass 3 column names not {len(column_names)}')
        self.column_names = column_names
        self.drug_name, self.protein_name, self.label_name = self.column_names

        # SMILES
        self.smiles_dataset = SMILESTokenizerDataset(
            self.smi_filepath,
            smiles_language=smiles_language,
            canonical=smiles_canonical,
            augment=smiles_augment,
            kekulize=smiles_kekulize,
            all_bonds_explicit=smiles_all_bonds_explicit,
            all_hs_explicit=smiles_all_hs_explicit,
            remove_bonddir=smiles_remove_bonddir,
            remove_chirality=smiles_remove_chirality,
            selfies=smiles_selfies,
            sanitize=smiles_sanitize,
            padding=smiles_padding,
            padding_length=smiles_padding_length,
            add_start_and_stop=smiles_add_start_and_stop,
            randomize=smiles_randomize,
            device=self.device,
            vocab_file=smiles_vocab_file,
            iterate_dataset=iterate_dataset,
            backend=self.backend,
        )
        # protein sequences
        self.protein_sequence_dataset = ProteinSequenceDataset(
            self.protein_filepath,
            protein_language=protein_language,
            amino_acid_dict=protein_amino_acid_dict,
            padding=protein_padding,
            padding_length=protein_padding_length,
            add_start_and_stop=protein_add_start_and_stop,
            augment_by_revert=protein_augment_by_revert,
            randomize=protein_randomize,
            device=self.device,
            iterate_dataset=iterate_dataset,
        )
        # drug affinity
        self.drug_affinity_dtype = drug_affinity_dtype
        self.drug_affinity_df = pd.read_csv(self.drug_affinity_filepath, index_col=0)
        # filter data based on the availability
        drug_mask = self.drug_affinity_df[self.drug_name].isin(
            set(self.smiles_dataset.keys())
        )
        sequence_mask = self.drug_affinity_df[self.protein_name].isin(
            set(self.protein_sequence_dataset.keys())
        )
        self.drug_affinity_df = self.drug_affinity_df.loc[drug_mask & sequence_mask]
        # to investigate missing ids per entity
        self.masks_df = pd.concat([drug_mask, sequence_mask], axis=1)
        self.masks_df.columns = [self.drug_name, self.protein_name]

        self.number_of_samples = len(self.drug_affinity_df)

    def __len__(self) -> int:
        "Total number of samples."
        return self.number_of_samples

    def __getitem__(self, index: int) -> DrugAffinityData:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            DrugAffinityData: a tuple containing three torch.Tensors,
                representing respectively: compound token indexes,
                protein sequence indexes and label for the current sample.
        """
        # drug affinity
        selected_sample = self.drug_affinity_df.iloc[index]
        affinity_tensor = torch.tensor(
            [selected_sample[self.label_name]],
            dtype=self.drug_affinity_dtype,
            device=self.device,
        )
        # SMILES
        token_indexes_tensor = self.smiles_dataset.get_item_from_key(
            selected_sample[self.drug_name]
        )
        # protein
        protein_sequence_tensor = self.protein_sequence_dataset.get_item_from_key(
            selected_sample[self.protein_name]
        )
        return token_indexes_tensor, protein_sequence_tensor, affinity_tensor



"""Testing ProteinFeatureLanguage."""
import os
import unittest

from pytoda.proteins.processing import (
    AA_FEAT,
    AA_PROPERTIES_NUM,
    BLOSUM62,
    BLOSUM62_NORM,
)
from pytoda.proteins.protein_feature_language import ProteinFeatureLanguage
from pytoda.tests.utils import TestFileContent


class TestProteinFeatureLanguage(unittest.TestCase):
    """Testing ProteinFeatureLanguage."""

    def test__update_max_token_sequence_length(self) -> None:
        """Test _update_max_token_sequence_length."""
        sequence = 'EGK'
        protein_language = ProteinFeatureLanguage(add_start_and_stop=False)
        self.assertEqual(protein_language.max_token_sequence_length, 0)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 3)
        protein_language = ProteinFeatureLanguage(add_start_and_stop=True)
        self.assertEqual(protein_language.max_token_sequence_length, 2)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 5)

        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=False, features='blosum'
        )
        self.assertEqual(protein_language.max_token_sequence_length, 0)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 3)
        protein_language = ProteinFeatureLanguage(add_start_and_stop=True)
        self.assertEqual(protein_language.max_token_sequence_length, 2)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 5)

    def test_add_file(self) -> None:
        """Test add_file"""
        content = os.linesep.join(['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4'])
        with TestFileContent(content) as a_test_file:
            protein_language = ProteinFeatureLanguage()
            protein_language.add_file(a_test_file.filename, index_col=1)
            self.assertEqual(protein_language.max_token_sequence_length, 7)

        # Test parsing of .fasta file
        content = r""">sp|Q6GZX0|005R_FRG3G Uncharacterized protein 005R OS=Frog virus 3 (isolate Goorha) OX=654924 GN=FV3-005R PE=4 SV=1
        MQNPLPEVMSPEHDKRTTTPMSKEANKFIRELDKKPGDLAVVSDFVKRNTGKRLPIGKRS
        NLYVRICDLSGTIYMGETFILESWEELYLPEPTKMEVLGTLESCCGIPPFPEWIVMVGED
        QCVYAYGDEEILLFAYSVKQLVEEGIQETGISYKYPDDISDVDEEVLQQDEEIQKIRKKT
        REFVDKDAQEFQDFLNSLDASLLS
        >sp|Q91G88|006L_IIV6 Putative KilA-N domain-containing protein 006L OS=Invertebrate iridescent virus 6 OX=176652 GN=IIV6-006L PE=3 SV=1
        MDSLNEVCYEQIKGTFYKGLFGDFPLIVDKKTGCFNATKLCVLGGKRFVDWNKTLRSKKL
        IQYYETRCDIKTESLLYEIKGDNNDEITKQITGTYLPKEFILDIASWISVEFYDKCNNII
        """

        with TestFileContent(content) as a_test_file:
            protein_language = ProteinFeatureLanguage(add_start_and_stop=False)
            protein_language.add_file(a_test_file.filename, file_type='.fasta')
            self.assertEqual(protein_language.max_token_sequence_length, 204)

    def test_sequence_to_token_indexes(self) -> None:
        """Test sequence_to_token_indexes."""
        sequence = 'CGX'
        protein_language = ProteinFeatureLanguage(add_start_and_stop=False)
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [BLOSUM62['C'], BLOSUM62['G'], BLOSUM62['X']],
        )
        protein_language = ProteinFeatureLanguage(add_start_and_stop=True)
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                BLOSUM62['<START>'],
                BLOSUM62['C'],
                BLOSUM62['G'],
                BLOSUM62['X'],
                BLOSUM62['<STOP>'],
            ],
        )
        # Other dictionary
        # Normed blosum
        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=False, features='blosum_norm'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [BLOSUM62_NORM['C'], BLOSUM62_NORM['G'], BLOSUM62_NORM['X']],
        )
        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=True, features='blosum_norm'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                BLOSUM62_NORM['<START>'],
                BLOSUM62_NORM['C'],
                BLOSUM62_NORM['G'],
                BLOSUM62_NORM['X'],
                BLOSUM62_NORM['<STOP>'],
            ],
        )
        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=False, features='binary_features'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [AA_PROPERTIES_NUM['C'], AA_PROPERTIES_NUM['G'], AA_PROPERTIES_NUM['X']],
        )
        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=True, features='float_features'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                AA_FEAT['<START>'],
                AA_FEAT['C'],
                AA_FEAT['G'],
                AA_FEAT['X'],
                AA_FEAT['<STOP>'],
            ],
        )
        # Test sequence with unknown token
        new_seq = 'CcX'
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(new_seq),
            [
                AA_FEAT['<START>'],
                AA_FEAT['C'],
                AA_FEAT['<UNK>'],
                AA_FEAT['X'],
                AA_FEAT['<STOP>'],
            ],
        )

    def test_token_indexes_to_sequence(self) -> None:
        """Test token_indexes_to_sequence."""
        sequence = 'CGX'
        protein_language = ProteinFeatureLanguage()
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CGX'
        )
        token_indexes = (
            [protein_language.token_to_index['<START>']]
            + token_indexes
            + [protein_language.token_to_index['<STOP>']]
        )
        protein_language = ProteinFeatureLanguage(add_start_and_stop=True)
        protein_language.add_sequence(sequence)
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CGX'
        )

        protein_language = ProteinFeatureLanguage(features='float_features')
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CGX'
        )
        token_indexes = (
            [protein_language.token_to_index['<START>']]
            + token_indexes
            + [protein_language.token_to_index['<STOP>']]
        )
        protein_language = ProteinFeatureLanguage(
            add_start_and_stop=True, features='float_features'
        )
        protein_language.add_sequence(sequence)
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CGX'
        )
        """
        NOTE: token_indexes_to_sequence for binary_features is impossible as
        multiple aa have the same encoding.
        """
        # Test whether code throws exception.
        protein_language = ProteinFeatureLanguage(features='binary_features')
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertRaises(
            Exception, protein_language.token_indexes_to_sequence, token_indexes
        )


if __name__ == '__main__':
    unittest.main()



"""KeyDataset for pandas DataFrames ."""
import numpy as np
import pandas as pd

from ..types import Hashable, Iterator
from .base_dataset import KeyDataset


class DataFrameDataset(KeyDataset):
    """
    Dataset of rows from pandas.DataFrame
    """

    def __init__(self, df: pd.DataFrame):
        super(DataFrameDataset).__init__()
        self.df = df

    def __len__(self) -> int:
        """Total number of samples."""
        return len(self.df)

    def __getitem__(self, index: int) -> np.array:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            np.array: a selected row of the CSV.
        """
        return self.df.iloc[index].values

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.df.index[index]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        # item will raise if not single value (deprecated in pandas)
        try:
            indexes = np.nonzero(self.df.index == key)[0]
            return indexes.item()
        except ValueError:
            if len(indexes) == 0:
                raise KeyError
            else:
                # key not unique, return first as ConcatKeyDataset
                return indexes[0]

    def get_item_from_key(self, key: Hashable) -> np.array:
        """Get item via key"""
        return self.df.loc[key, :].values

    def keys(self) -> Iterator:
        """Iterator over index of dataframe."""
        return iter(self.df.index)



"""Implementation of DrugSensitivityDataset."""
import pandas as pd
import torch
from torch.utils.data import Dataset

from ..smiles.smiles_language import SMILESLanguage
from ..types import DrugSensitivityData, GeneList, Iterable, Tuple
from .gene_expression_dataset import GeneExpressionDataset
from .smiles_dataset import SMILESTokenizerDataset


class DrugSensitivityDataset(Dataset):
    """
    Drug sensitivity dataset implementation.
    """

    def __init__(
        self,
        drug_sensitivity_filepath: str,
        smi_filepath: str,
        gene_expression_filepath: str,
        column_names: Tuple[str] = ['drug', 'cell_line', 'IC50'],
        drug_sensitivity_dtype: torch.dtype = torch.float,
        drug_sensitivity_min_max: bool = True,
        drug_sensitivity_processing_parameters: dict = {},
        smiles_language: SMILESLanguage = None,
        padding: bool = True,
        padding_length: int = None,
        add_start_and_stop: bool = False,
        augment: bool = False,
        canonical: bool = False,
        kekulize: bool = False,
        all_bonds_explicit: bool = False,
        all_hs_explicit: bool = False,
        randomize: bool = False,
        remove_bonddir: bool = False,
        remove_chirality: bool = False,
        selfies: bool = False,
        sanitize: bool = True,
        vocab_file: str = None,
        iterate_dataset: bool = True,
        gene_list: GeneList = None,
        gene_expression_standardize: bool = True,
        gene_expression_min_max: bool = False,
        gene_expression_processing_parameters: dict = {},
        gene_expression_dtype: torch.dtype = torch.float,
        gene_expression_kwargs: dict = {},
        device: torch.device = (
            torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        ),
        backend: str = 'eager',
    ) -> None:
        """
        Initialize a drug sensitivity dataset.

        Args:
            drug_sensitivity_filepath (str): path to drug sensitivity .csv file.
                Currently, the only supported format is .csv, with an index and three
                header columns named as specified in the column_names argument.
            smi_filepath (str): path to .smi file.
            gene_expression_filepath (str): path to gene expression .csv file.
                Currently, the only supported format is .csv,
                with an index and header columns containing gene names.
            column_names (Tuple[str]): Names of columns in data files to retrieve
                labels, ligands and protein name respectively.
                Defaults to ['drug', 'cell_line', 'IC50'].
            drug_sensitivity_dtype (torch.dtype): drug sensitivity data type.
                Defaults to torch.float.
            drug_sensitivity_min_max (bool): min-max scale drug sensitivity
                data. Defaults to True.
            drug_sensitivity_processing_parameters (dict): transformation
                parameters for drug sensitivity data, e.g. for min-max scaling.
                Defaults to {}.
            smiles_language (SMILESLanguage): a smiles language.
                Defaults to None.
            padding (bool): pad sequences to longest in the smiles language.
                Defaults to True.
            padding_length (int): manually sets number of applied paddings,
                applies only if padding is True. Defaults to None.
            add_start_and_stop (bool): add start and stop token indexes.
                Defaults to False.
            canonical (bool): performs canonicalization of SMILES (one
                original string for one molecule), if True, then other
                transformations (augment etc, see below) do not apply
            augment (bool): perform SMILES augmentation. Defaults to False.
            kekulize (bool): kekulizes SMILES (implicit aromaticity only).
                Defaults to False.
            all_bonds_explicit (bool): Makes all bonds explicit. Defaults to
                False, only applies if kekulize = True.
            all_hs_explicit (bool): Makes all hydrogens explicit. Defaults to
                False, only applies if kekulize = True.
            randomize (bool): perform a true randomization of SMILES tokens.
                Defaults to False.
            remove_bonddir (bool): Remove directional info of bonds.
                Defaults to False.
            remove_chirality (bool): Remove chirality information.
                Defaults to False.
            selfies (bool): Whether selfies is used instead of smiles, defaults
                to False.
            sanitize (bool): RDKit sanitization of the molecule.
                Defaults to True.
            vocab_file (str): Optional .json to load vocabulary. Tries to load
                metadata if `iterate_dataset` is False. Defaults to None.
            iterate_dataset (bool): whether to go through all SMILES in the
                dataset to extend/build vocab, find longest sequence, and
                checks the passed padding length if applicable. Defaults to
                True.
            gene_list (GeneList): a list of genes.
            gene_expression_standardize (bool): perform gene expression
                data standardization. Defaults to True.
            gene_expression_min_max (bool): perform min-max scaling on gene
                expression data. Defaults to False.
            gene_expression_processing_parameters (dict): transformation
                parameters for gene expression, e.g. for min-max scaling.
                Defaults to {}.
            gene_expression_dtype (torch.dtype): gene expression data type.
                Defaults to torch.float.
            gene_expression_kwargs (dict): additional parameters for
                GeneExpressionDataset.
            device (torch.device): device where the tensors are stored.
                Defaults to gpu, if available.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
                Note that at the moment only the gene expression and the
                smiles datasets implement both backends. The drug sensitivity
                data are loaded in memory.
        """
        Dataset.__init__(self)
        self.drug_sensitivity_filepath = drug_sensitivity_filepath
        self.smi_filepath = smi_filepath
        self.gene_expression_filepath = gene_expression_filepath
        # device
        self.device = device
        # backend
        self.backend = backend

        if not isinstance(column_names, Iterable):
            raise TypeError(f'Column names was {type(column_names)}, not Iterable.')
        if not len(column_names) == 3:
            raise ValueError(f'Please pass 3 column names not {len(column_names)}')
        self.column_names = column_names
        self.drug_name, self.cell_name, self.label_name = self.column_names

        # SMILES
        self.smiles_dataset = SMILESTokenizerDataset(
            self.smi_filepath,
            smiles_language=smiles_language,
            augment=augment,
            canonical=canonical,
            kekulize=kekulize,
            all_bonds_explicit=all_bonds_explicit,
            all_hs_explicit=all_hs_explicit,
            remove_bonddir=remove_bonddir,
            remove_chirality=remove_chirality,
            selfies=selfies,
            sanitize=sanitize,
            randomize=randomize,
            padding=padding,
            padding_length=padding_length,
            add_start_and_stop=add_start_and_stop,
            device=self.device,
            vocab_file=vocab_file,
            iterate_dataset=iterate_dataset,
            backend=self.backend,
        )
        # gene expression
        self.gene_expression_dataset = GeneExpressionDataset(
            self.gene_expression_filepath,
            gene_list=gene_list,
            standardize=gene_expression_standardize,
            min_max=gene_expression_min_max,
            processing_parameters=gene_expression_processing_parameters,
            dtype=gene_expression_dtype,
            device=self.device,
            backend=self.backend,
            index_col=0,
            **gene_expression_kwargs,
        )
        # drug sensitivity
        self.drug_sensitivity_dtype = drug_sensitivity_dtype
        self.drug_sensitivity_min_max = drug_sensitivity_min_max
        self.drug_sensitivity_processing_parameters = (
            drug_sensitivity_processing_parameters
        )
        self.drug_sensitivity_df = pd.read_csv(
            self.drug_sensitivity_filepath, index_col=0
        )
        # filter data based on the availability
        drug_mask = self.drug_sensitivity_df[self.drug_name].isin(
            set(self.smiles_dataset.keys())
        )
        profile_mask = self.drug_sensitivity_df[self.cell_name].isin(
            set(self.gene_expression_dataset.keys())
        )
        self.drug_sensitivity_df = self.drug_sensitivity_df.loc[
            drug_mask & profile_mask
        ]

        # to investigate missing ids per entity
        self.masks_df = pd.concat([drug_mask, profile_mask], axis=1)
        self.masks_df.columns = [self.drug_name, self.cell_name]

        self.number_of_samples = len(self.drug_sensitivity_df)

        # NOTE: optional min-max scaling
        if self.drug_sensitivity_min_max:
            minimum = self.drug_sensitivity_processing_parameters.get(
                'min', self.drug_sensitivity_df[self.label_name].min()
            )
            maximum = self.drug_sensitivity_processing_parameters.get(
                'max', self.drug_sensitivity_df[self.label_name].max()
            )
            self.drug_sensitivity_df[self.label_name] = (
                self.drug_sensitivity_df[self.label_name] - minimum
            ) / (maximum - minimum)
            self.drug_sensitivity_processing_parameters = {
                'processing': 'min_max',
                'parameters': {'min': minimum, 'max': maximum},
            }

    def __len__(self) -> int:
        "Total number of samples."
        return self.number_of_samples

    def __getitem__(self, index: int) -> DrugSensitivityData:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            DrugSensitivityData: a tuple containing three torch.Tensors,
                representing respectively: compound token indexes,
                gene expression values and IC50 for the current sample.
        """
        # drug sensitivity
        selected_sample = self.drug_sensitivity_df.iloc[index]
        ic50_tensor = torch.tensor(
            [selected_sample[self.label_name]],
            dtype=self.drug_sensitivity_dtype,
            device=self.device,
        )
        # SMILES
        token_indexes_tensor = self.smiles_dataset.get_item_from_key(
            selected_sample[self.drug_name]
        )
        # gene_expression
        gene_expression_tensor = self.gene_expression_dataset.get_item_from_key(
            selected_sample[self.cell_name]
        )
        return token_indexes_tensor, gene_expression_tensor, ic50_tensor



"""Testing PolymerTokenizer."""
import os
import tempfile
import unittest

from pytoda.smiles.polymer_language import PolymerTokenizer
from pytoda.smiles.processing import split_selfies
from pytoda.smiles.transforms import Selfies
from pytoda.tests.utils import TestFileContent


class TestPolymerTokenizer(unittest.TestCase):
    """Testing PolymerTokenizer."""

    def test__update_max_token_sequence_length(self) -> None:
        """Test _update_max_token_sequence_length."""
        smiles = 'CCO'
        entities = ['Initiator', 'Monomer', 'Catalyst']
        polymer_language = PolymerTokenizer(entity_names=entities)
        self.assertEqual(polymer_language.max_token_sequence_length, 0)
        polymer_language.add_smiles(smiles)
        self.assertEqual(polymer_language.max_token_sequence_length, 5)

    def test__update_language_dictionaries_with_tokens(self) -> None:
        """Test _update_language_dictionaries_with_tokens."""
        smiles = 'CCO'
        entities = ['Initiator', 'Monomer', 'Catalyst']
        polymer_language = PolymerTokenizer(entity_names=entities)
        polymer_language._update_language_dictionaries_with_tokens(
            polymer_language.smiles_tokenizer(smiles)
        )
        self.assertTrue(
            'C' in polymer_language.token_to_index
            and 'O' in polymer_language.token_to_index
        )
        self.assertEqual(polymer_language.number_of_tokens, 43)

    def test_add_smis(self) -> None:
        """Test add_smis."""
        content = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        with TestFileContent(content) as a_test_file:
            with TestFileContent(content) as another_test_file:
                entities = ['Initiator', 'Monomer', 'Catalyst']
                polymer_language = PolymerTokenizer(entity_names=entities)
                polymer_language.add_smis(
                    [a_test_file.filename, another_test_file.filename]
                )
                self.assertEqual(polymer_language.number_of_tokens, 45)

    def test_add_smi(self) -> None:
        """Test add_smi."""
        content = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        with TestFileContent(content) as test_file:
            entities = ['Initiator', 'Monomer']
            polymer_language = PolymerTokenizer(entity_names=entities)
            polymer_language.add_smi(test_file.filename)
            self.assertEqual(polymer_language.number_of_tokens, 43)

    def test_add_smiles(self) -> None:
        """Test add_smiles."""
        smiles = 'CCO'
        entities = ['Initiator', 'Monomer']
        polymer_language = PolymerTokenizer(entity_names=entities)
        polymer_language.add_smiles(smiles)
        self.assertEqual(polymer_language.number_of_tokens, 41)

    def test_smiles_to_token_indexes(self) -> None:
        """Test smiles_to_token_indexes."""

        smiles = 'CCO'
        entities = ['Initiator', 'Monomer', 'Catalyst']
        polymer_language = PolymerTokenizer(entity_names=entities)
        polymer_language.add_smiles(smiles)
        token_indexes = [polymer_language.token_to_index[token] for token in smiles]
        polymer_language.update_entity('monomer')
        self.assertListEqual(
            list(polymer_language.smiles_to_token_indexes(smiles)),
            [polymer_language.token_to_index['<MONOMER_START>']]
            + token_indexes
            + [polymer_language.token_to_index['<MONOMER_STOP>']],
        )
        polymer_language.update_entity('catalyst')
        self.assertListEqual(
            list(polymer_language.smiles_to_token_indexes(smiles)),
            [polymer_language.token_to_index['<CATALYST_START>']]
            + token_indexes
            + [polymer_language.token_to_index['<CATALYST_STOP>']],
        )

        # SELFIES
        polymer_language = PolymerTokenizer(
            entity_names=entities, smiles_tokenizer=split_selfies
        )
        transform = Selfies()
        selfies = transform(smiles)
        polymer_language.add_smiles(selfies)
        token_indexes = [
            polymer_language.token_to_index[token] for token in ['[C]', '[C]', '[O]']
        ]
        polymer_language.update_entity('monomer')
        self.assertListEqual(
            list(polymer_language.smiles_to_token_indexes(selfies)),
            [polymer_language.token_to_index['<MONOMER_START>']]
            + token_indexes
            + [polymer_language.token_to_index['<MONOMER_STOP>']],
        )

    def test_token_indexes_to_smiles(self) -> None:
        """Test token_indexes_to_smiles."""
        smiles = 'CCO'
        entities = ['Initiator', 'Monomer', 'Catalyst']
        polymer_language = PolymerTokenizer(entity_names=entities)

        polymer_language.add_smiles(smiles)
        token_indexes = [polymer_language.token_to_index[token] for token in smiles]
        self.assertEqual(polymer_language.token_indexes_to_smiles(token_indexes), 'CCO')
        token_indexes = (
            [polymer_language.token_to_index['<MONOMER_START>']]
            + token_indexes
            + [polymer_language.token_to_index['<MONOMER_STOP>']]
        )
        self.assertEqual(polymer_language.token_indexes_to_smiles(token_indexes), 'CCO')

    def test_vocab_roundtrip(self):
        smiles = 'CCO'
        entities = ['Initiator', 'Monomer', 'Catalyst']
        source_language = PolymerTokenizer(entity_names=entities)
        source_language.add_smiles(smiles)
        # to test
        vocab = source_language.token_to_index
        vocab_ = source_language.index_to_token
        max_len = source_language.max_token_sequence_length
        count = source_language.token_count
        total = source_language.number_of_tokens

        # just vocab
        with tempfile.TemporaryDirectory() as tempdir:
            source_language.save_vocabulary(tempdir)

            polymer_language = PolymerTokenizer(entity_names=entities)
            polymer_language.load_vocabulary(tempdir)
        self.assertDictEqual(vocab, polymer_language.token_to_index)
        self.assertDictEqual(vocab_, polymer_language.index_to_token)

        # pretrained
        with tempfile.TemporaryDirectory() as tempdir:
            source_language.save_pretrained(tempdir)

            polymer_language = PolymerTokenizer.from_pretrained(tempdir)
        self.assertDictEqual(vocab, polymer_language.token_to_index)
        self.assertDictEqual(vocab_, polymer_language.index_to_token)

        self.assertEqual(max_len, polymer_language.max_token_sequence_length)
        self.assertDictEqual(count, polymer_language.token_count)
        self.assertEqual(total, polymer_language.number_of_tokens)
        self.assertEqual(entities, polymer_language.entities)


if __name__ == '__main__':
    unittest.main()



"""Testing SMILES transforms."""
import unittest

import numpy as np
import torch

from pytoda.smiles.smiles_language import SMILESTokenizer
from pytoda.smiles.transforms import AugmentTensor, Kekulize, NotKekulize, RemoveIsomery


class TestTransforms(unittest.TestCase):
    """Testing transforms."""

    def test_kekulize(self):
        sanitize_opts = [True, False]
        for sanitize in sanitize_opts:
            self._test_kekulize(sanitize)

    def _test_kekulize(self, sanitize) -> None:
        """Test Kekulize."""
        for smiles, ground_truth in [
            ('c1cnoc1', 'C1=CON=C1'),
            ('[O-][n+]1ccccc1S', '[O-][N+]1=CC=CC=C1S'),
            ('c1snnc1-c1ccccn1', 'C1=C(C2=CC=CC=N2)N=NS1'),
        ]:
            transform = Kekulize(
                all_bonds_explicit=False, all_hs_explicit=False, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', 'C1=C-O-N=C-1'),
            ('[O-][n+]1ccccc1S', '[O-]-[N+]1=C-C=C-C=C-1-S'),
            ('c1snnc1-c1ccccn1', 'C1=C(-C2=C-C=C-C=N-2)-N=N-S-1'),
        ]:
            transform = Kekulize(
                all_bonds_explicit=True, all_hs_explicit=False, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', '[CH]1=[CH][O][N]=[CH]1'),
            ('[O-][n+]1ccccc1S', '[O-][N+]1=[CH][CH]=[CH][CH]=[C]1[SH]'),
            ('c1snnc1-c1ccccn1', '[CH]1=[C]([C]2=[CH][CH]=[CH][CH]=[N]2)[N]=[N][S]1'),
        ]:
            transform = Kekulize(
                all_bonds_explicit=False, all_hs_explicit=True, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', '[CH]1=[CH]-[O]-[N]=[CH]-1'),
            ('[O-][n+]1ccccc1S', '[O-]-[N+]1=[CH]-[CH]=[CH]-[CH]=[C]-1-[SH]'),
            (
                'c1snnc1-c1ccccn1',
                '[CH]1=[C](-[C]2=[CH]-[CH]=[CH]-[CH]=[N]-2)-[N]=[N]-[S]-1',
            ),
        ]:
            transform = Kekulize(
                all_bonds_explicit=True, all_hs_explicit=True, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

    def test_non_kekulize(self) -> None:
        sanitize_opts = [True, False]
        for sanitize in sanitize_opts:
            self._test_non_kekulize(sanitize)

    def _test_non_kekulize(self, sanitize) -> None:
        """Test NotKekulize."""
        for smiles, ground_truth in [
            ('c1cnoc1', 'c1cnoc1'),
            ('[O-][n+]1ccccc1S', '[O-][n+]1ccccc1S'),
            ('c1snnc1-c1ccccn1', 'c1snnc1-c1ccccn1'),
        ]:
            transform = NotKekulize(
                all_bonds_explicit=False, all_hs_explicit=False, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', 'c1:c:n:o:c:1'),
            ('[O-][n+]1ccccc1S', '[O-]-[n+]1:c:c:c:c:c:1-S'),
            ('c1snnc1-c1ccccn1', 'c1:s:n:n:c:1-c1:c:c:c:c:n:1'),
        ]:
            transform = NotKekulize(
                all_bonds_explicit=True, all_hs_explicit=False, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', '[cH]1[cH][n][o][cH]1'),
            ('[O-][n+]1ccccc1S', '[O-][n+]1[cH][cH][cH][cH][c]1[SH]'),
            ('c1snnc1-c1ccccn1', '[cH]1[s][n][n][c]1-[c]1[cH][cH][cH][cH][n]1'),
        ]:
            transform = NotKekulize(
                all_bonds_explicit=False, all_hs_explicit=True, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

        for smiles, ground_truth in [
            ('c1cnoc1', '[cH]1:[cH]:[n]:[o]:[cH]:1'),
            ('[O-][n+]1ccccc1S', '[O-]-[n+]1:[cH]:[cH]:[cH]:[cH]:[c]:1-[SH]'),
            (
                'c1snnc1-c1ccccn1',
                '[cH]1:[s]:[n]:[n]:[c]:1-[c]1:[cH]:[cH]:[cH]:[cH]:[n]:1',
            ),
        ]:
            transform = NotKekulize(
                all_bonds_explicit=True, all_hs_explicit=True, sanitize=sanitize
            )
            self.assertEqual(transform(smiles), ground_truth)

    def test_remove_isomery(self) -> None:
        """Test RemoveIsomery."""

        for bonddir, chirality, smiles, ground_truth in zip(
            [False, False, True, True],
            [False, True, False, True],
            4 * ['c1ccc(/C=C/[C@H](C)O)cc1'],
            [
                'c1ccc(/C=C/[C@H](C)O)cc1',
                'c1ccc(/C=C/C(C)O)cc1',
                'c1ccc(C=C[C@H](C)O)cc1',
                'c1ccc(C=CC(C)O)cc1',
            ],
        ):
            transform = RemoveIsomery(
                bonddir=bonddir, chirality=chirality, sanitize=True
            )
            self.assertEqual(transform(smiles), ground_truth)

    def test_augment_tensor(self) -> None:
        sanitize_opts = [True, False]
        for sanitize in sanitize_opts:
            for canonical in [False, True]:
                self._test_augment_tensor(sanitize, canonical)

    def _test_augment_tensor(self, sanitize, canonical) -> None:
        """Test AugmentTensor."""
        smiles = 'NCCS'
        smiles_language = SMILESTokenizer(
            add_start_and_stop=True, padding=False, canonical=canonical
        )
        smiles_language.add_smiles(smiles)

        np.random.seed(0)
        transform = AugmentTensor(smiles_language, sanitize=sanitize)

        def _transform_without_encoding(smile):
            """Encode SMILES by skippping transforms in smiles language. This is
            needed because the new SMILESTokenizer.smiles_to_token_indexes also
            implements SMILES transforms. This could silently canonicalize all SMILES
            if canonical=True in the smiles language object. Hence it's inappropriate
            to generate the ground truth and replace by this method."""
            return smiles_language.transform_encoding(
                [
                    smiles_language.token_to_index[t]
                    for t in smiles_language.smiles_tokenizer(smile)
                ]
            )

        token_indexes_tensor = _transform_without_encoding(smiles)

        for augmented_smile in ['C(S)CN', 'NCCS', 'SCCN', 'C(N)CS', 'C(CS)N']:
            ground_truth = _transform_without_encoding(augmented_smile)
            self.assertSequenceEqual(
                transform(token_indexes_tensor).tolist(), ground_truth.tolist()
            )

        # Now test calling with a tensor of several SMILES
        # Include the padding of the sequence (right padding)
        pl = 5  # padding_length
        single_smiles_tensor = torch.unsqueeze(
            torch.nn.functional.pad(
                token_indexes_tensor, (0, pl), value=smiles_language.padding_index
            ),
            0,
        )
        seq_len = single_smiles_tensor.shape[1]  # sequence_length
        multi_smiles_tensor = torch.cat([single_smiles_tensor] * 5)
        np.random.seed(0)
        augmented = transform(multi_smiles_tensor)

        for ind, augmented_smile in enumerate(
            ['C(S)CN', 'NCCS', 'SCCN', 'C(N)CS', 'C(CS)N']
        ):
            ground_truth = _transform_without_encoding(augmented_smile)
            ground_truth = torch.nn.functional.pad(
                ground_truth,
                pad=(0, seq_len - len(ground_truth)),
                value=smiles_language.padding_index,
            )
            self.assertSequenceEqual(augmented[ind].tolist(), ground_truth.tolist())


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Tuple, Union

import torch
from torch.random import fork_rng
from torch.utils.data import Dataset

from pytoda.types import Any, Tensor


class StochasticItems:
    """Sample an item from a distribution on the fly on indexing.

    Args:
        distribution (torch.distributions.distribution.Distribution): An instance
            of the torch distribution class to sample from. For example, for
            loc = torch.tensor(0.0,device=device), and scale=torch.tensor(1.0,device=device),
            torch.distributions.normal.Normal(loc,scale), so that
            calling .sample() would return an item from this distribution on the specified device.
            NOTE: The arguments to the distribution should be tensors on the desired device.
            This ensure that samples are generated on this device and helps in avoiding
            an overhead in sending each sampled item to device.
        shape (torch.Size): The desired shape of each item.
        device (torch.device): Device to send the tensor to. If the tensor is already
            on device, then the .to() method returns self (no-ops).
    """

    def __init__(
        self,
        distribution: torch.distributions.distribution.Distribution,
        shape: Union[torch.Size, Tuple[int]],
        device: torch.device,
    ):

        self.distribution = distribution
        self.shape = shape
        self.device = device

        # check if distribution arguments are on device
        devices = []
        for key in distribution.arg_constraints:
            devices.append(getattr(distribution, key).device.type)

        args_device = set(devices)

        if len(args_device) > 1:
            raise RuntimeError(
                f"Expected all tensors to be on the same device, but found {args_device} instead."
            )

        elif args_device != {device.type}:
            raise RuntimeWarning(
                f"Expected arguments to be on {device}, but they are on {args_device} instead. This will cause a data transfer overhead."
            )

    def __getitem__(self, index: Any) -> Tensor:
        """Samples an item.

        Args:
            index (Any): is ignored.

        Returns:
            Tensor: sampled from distribution with given shape.
        """

        return self.distribution.sample(self.shape).to(self.device)


class DistributionalDataset(Dataset):
    """Generates samples from a specified distribution."""

    def __init__(
        self,
        dataset_size: int,
        item_shape: Tuple[int],
        distribution_function: torch.distributions.distribution.Distribution,
        seed: Optional[int] = None,
        device: torch.device = (
            torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        ),
    ) -> None:
        """Dataset of synthetic samples from a specified distribution with given shape.

        Args:
            dataset_size (int): Number of items to generate (N).
            item_shape (Tuple[int]): The shape of each item tensor returned on
                indexing the dataset. For example for 2D items with timeseries of
                3 timesteps and 5 features: (3, 5)
            distribution_function (torch.distributions.distribution.Distribution):
                An instance of the distribution class from which individual data
                items can be sampled by calling the .sample() method.
                This can either be an object that is directly initialised using a method
                from torch.distributions, such as, torch.distributions.normal.Normal(loc=0.0,scale=1.0),
                or from a factory using a keyword, for example,
                DISTRIBUTION_FUNCTION_FACTORY['normal](loc=0.0, scale=1.0)
                is a valid argument since the factory (found in utils.factories.py)
                initialises the distribution class object based on a string keyword
                and passes the relevant arguments to that object.
            seed (Optional[int]): If passed, all items are generated once with
                this seed (using a local RNG only). Defaults to None, where
                individual items are generated when the DistributionalDataset is
                indexed (using the global RNG).
            device (torch.device): Device where the tensors are stored.
                Defaults to gpu, if available.
        """

        super(DistributionalDataset, self).__init__()

        self.dataset_size = dataset_size
        self.item_shape = item_shape
        self.seed = seed
        self.device = device

        self.data_sampler = distribution_function

        if self.seed:
            # Eager dataset creation
            with fork_rng():
                torch.manual_seed(seed)
                torch.cuda.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)

                self.datasource = self.data_sampler.sample((dataset_size, *item_shape))

            self.datasource = self.datasource.to(device)

        else:
            # get sampled item on indexing
            self.datasource = StochasticItems(
                self.data_sampler, self.item_shape, self.device
            )

        # copy data to device

    def __len__(self) -> int:
        """Gets length of dataset.

        Returns:
            int: Length of the set being sampled.
        """
        return self.dataset_size

    def __getitem__(self, index: int) -> torch.Tensor:
        """Generates a single set sample.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            torch.Tensor: Tensor containing elements with shape T x H, i.e.
                [self.dataset_depth, self.dataset_dim].
        """

        return self.datasource[index]



"""Testing ProteinLanguage."""
import os
import unittest

from pytoda.proteins.processing import (
    HUMAN_KINASE_ALIGNMENT_VOCAB,
    IUPAC_VOCAB,
    UNIREP_VOCAB,
)
from pytoda.proteins.protein_language import ProteinLanguage
from pytoda.tests.utils import TestFileContent


class TestProteinLanguage(unittest.TestCase):
    """Testing ProteinLanguage."""

    def test__update_max_token_sequence_length(self) -> None:
        """Test _update_max_token_sequence_length."""
        sequence = 'EGK'
        protein_language = ProteinLanguage(add_start_and_stop=False)
        self.assertEqual(protein_language.max_token_sequence_length, 0)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 3)
        protein_language = ProteinLanguage(add_start_and_stop=True)
        self.assertEqual(protein_language.max_token_sequence_length, 2)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 5)

        protein_language = ProteinLanguage(
            add_start_and_stop=False, amino_acid_dict='unirep'
        )
        self.assertEqual(protein_language.max_token_sequence_length, 0)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 3)
        protein_language = ProteinLanguage(add_start_and_stop=True)
        self.assertEqual(protein_language.max_token_sequence_length, 2)
        protein_language.add_sequence(sequence)
        self.assertEqual(protein_language.max_token_sequence_length, 5)

    def test_add_file(self) -> None:
        """Test add_file"""
        content = os.linesep.join(['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4'])
        with TestFileContent(content) as a_test_file:
            protein_language = ProteinLanguage()
            protein_language.add_file(a_test_file.filename, index_col=1)
            self.assertEqual(protein_language.max_token_sequence_length, 7)

        # Test parsing of .fasta file
        content = r""">sp|Q6GZX0|005R_FRG3G Uncharacterized protein 005R OS=Frog virus 3 (isolate Goorha) OX=654924 GN=FV3-005R PE=4 SV=1
        MQNPLPEVMSPEHDKRTTTPMSKEANKFIRELDKKPGDLAVVSDFVKRNTGKRLPIGKRS
        NLYVRICDLSGTIYMGETFILESWEELYLPEPTKMEVLGTLESCCGIPPFPEWIVMVGED
        QCVYAYGDEEILLFAYSVKQLVEEGIQETGISYKYPDDISDVDEEVLQQDEEIQKIRKKT
        REFVDKDAQEFQDFLNSLDASLLS
        >sp|Q91G88|006L_IIV6 Putative KilA-N domain-containing protein 006L OS=Invertebrate iridescent virus 6 OX=176652 GN=IIV6-006L PE=3 SV=1
        MDSLNEVCYEQIKGTFYKGLFGDFPLIVDKKTGCFNATKLCVLGGKRFVDWNKTLRSKKL
        IQYYETRCDIKTESLLYEIKGDNNDEITKQITGTYLPKEFILDIASWISVEFYDKCNNII
        """

        with TestFileContent(content) as a_test_file:
            protein_language = ProteinLanguage(add_start_and_stop=False)
            protein_language.add_file(a_test_file.filename, file_type='.fasta')
            self.assertEqual(protein_language.max_token_sequence_length, 204)

    def test_sequence_to_token_indexes(self) -> None:
        """Test sequence_to_token_indexes."""
        sequence = 'CCO'
        protein_language = ProteinLanguage(add_start_and_stop=False)
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [IUPAC_VOCAB['C'], IUPAC_VOCAB['C'], IUPAC_VOCAB['O']],
        )
        protein_language = ProteinLanguage(add_start_and_stop=True)
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                IUPAC_VOCAB['<START>'],
                IUPAC_VOCAB['C'],
                IUPAC_VOCAB['C'],
                IUPAC_VOCAB['O'],
                IUPAC_VOCAB['<STOP>'],
            ],
        )

        # UniRep
        protein_language = ProteinLanguage(
            add_start_and_stop=False, amino_acid_dict='unirep'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [UNIREP_VOCAB['C'], UNIREP_VOCAB['C'], UNIREP_VOCAB['O']],
        )
        protein_language = ProteinLanguage(
            add_start_and_stop=True, amino_acid_dict='unirep'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                UNIREP_VOCAB['<START>'],
                UNIREP_VOCAB['C'],
                UNIREP_VOCAB['C'],
                UNIREP_VOCAB['O'],
                UNIREP_VOCAB['<STOP>'],
            ],
        )

        # human kinase alignment
        sequence = 'AC-C'
        protein_language = ProteinLanguage(
            add_start_and_stop=False, amino_acid_dict='human-kinase-alignment'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                HUMAN_KINASE_ALIGNMENT_VOCAB['A'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['C'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['-'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['C'],
            ],
        )
        protein_language = ProteinLanguage(
            add_start_and_stop=True, amino_acid_dict='human-kinase-alignment'
        )
        protein_language.add_sequence(sequence)
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(sequence),
            [
                HUMAN_KINASE_ALIGNMENT_VOCAB['<START>'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['A'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['C'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['-'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['C'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['<STOP>'],
            ],
        )

        # Test encoding an unknown character
        new_seq = 'Aq-C'
        self.assertListEqual(
            protein_language.sequence_to_token_indexes(new_seq),
            [
                HUMAN_KINASE_ALIGNMENT_VOCAB['<START>'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['A'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['<UNK>'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['-'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['C'],
                HUMAN_KINASE_ALIGNMENT_VOCAB['<STOP>'],
            ],
        )

    def test_token_indexes_to_sequence(self) -> None:
        """Test token_indexes_to_sequence."""
        sequence = 'CCO'
        protein_language = ProteinLanguage()
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CCO'
        )
        token_indexes = (
            [protein_language.token_to_index['<START>']]
            + token_indexes
            + [protein_language.token_to_index['<STOP>']]
        )
        protein_language = ProteinLanguage(add_start_and_stop=True)
        protein_language.add_sequence(sequence)
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CCO'
        )

        # UniRep
        protein_language = ProteinLanguage(amino_acid_dict='unirep')
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CCO'
        )
        token_indexes = (
            [protein_language.token_to_index['<START>']]
            + token_indexes
            + [protein_language.token_to_index['<STOP>']]
        )
        protein_language = ProteinLanguage(
            add_start_and_stop=True, amino_acid_dict='unirep'
        )
        protein_language.add_sequence(sequence)
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), 'CCO'
        )

        # human kinase alignment
        sequence = 'AC-C'
        protein_language = ProteinLanguage(amino_acid_dict='human-kinase-alignment')
        protein_language.add_sequence(sequence)
        token_indexes = [protein_language.token_to_index[token] for token in sequence]
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), sequence
        )
        token_indexes = (
            [protein_language.token_to_index['<START>']]
            + token_indexes
            + [protein_language.token_to_index['<STOP>']]
        )
        protein_language = ProteinLanguage(
            add_start_and_stop=True, amino_acid_dict='human-kinase-alignment'
        )
        protein_language.add_sequence(sequence)
        self.assertEqual(
            protein_language.token_indexes_to_sequence(token_indexes), sequence
        )


if __name__ == '__main__':
    unittest.main()



"""Testing DistributionalDataset."""
import unittest

import torch

from pytoda.datasets import DistributionalDataset
from pytoda.datasets.utils.factories import DISTRIBUTION_FUNCTION_FACTORY

distribution_types = ['normal', 'uniform']
distribution_args = [{'loc': 0.0, 'scale': 1.0}, {'low': 0.0, 'high': 1.0}]
dataset_sizes = [100, 10000]
item_shapes = [(1, 16), (10, 16)]
seeds = [None, 1, 42]


class TestDistributionalDataset(unittest.TestCase):
    """Test DistributionalDataset class."""

    def test__len__(self) -> None:
        """Test __len__."""

        for dist_type, dist_args in zip(distribution_types, distribution_args):
            distribution_function = DISTRIBUTION_FUNCTION_FACTORY[dist_type](
                **dist_args
            )
            for size in dataset_sizes:
                for shape in item_shapes:
                    for seed in seeds:
                        dataset = DistributionalDataset(
                            size,
                            shape,
                            distribution_function,
                            seed=seed,
                        )
                        self.assertEqual(len(dataset), size)

    def test__getitem__(self) -> None:
        """Test __getitem__."""

        for dist_type, dist_args in zip(distribution_types, distribution_args):
            distribution_function = DISTRIBUTION_FUNCTION_FACTORY[dist_type](
                **dist_args
            )
            for size in dataset_sizes:
                for shape in item_shapes:
                    for seed in seeds:
                        dataset = DistributionalDataset(
                            size,
                            shape,
                            distribution_function,
                            seed=seed,
                        )
                        sample1_1 = dataset[42]
                        sample1_2 = dataset[42]

                        # Test shapes
                        self.assertEqual(sample1_1.shape, sample1_2.shape)
                        self.assertEqual(sample1_1.shape, shape)
                        # Test content
                        if seed is None:
                            self.assertFalse(torch.equal(sample1_1, sample1_2))
                        else:
                            self.assertTrue(torch.equal(sample1_1, sample1_2))


if __name__ == '__main__':
    unittest.main()



"""Utilities for file handling."""
import os
from itertools import repeat, takewhile
from typing import Sequence

import pandas as pd


def count_file_lines(filepath: str, buffer_size: int = 1024 * 1024) -> int:
    """
    Count lines in a file without persisting it in memory.

    Args:
        filepath (str): path to the file.
        buffer_size (int): size of the buffer.

    Returns:
        int: Number of lines in the file.
    """
    number_of_lines = 0
    new_line = os.linesep.encode()
    with open(filepath, 'rb') as fp:
        raw_fp = fp.raw
        previous_buffer = None
        for buffer in takewhile(
            lambda x: x, (raw_fp.read(buffer_size) for _ in repeat(None))
        ):
            number_of_lines += buffer.count(new_line)
            previous_buffer = buffer
        number_of_lines += int(not previous_buffer.endswith(new_line))
    return number_of_lines


def read_smi(
    filepath: str,
    chunk_size: int = None,
    index_col: int = 1,
    names: Sequence[str] = ['SMILES'],
    header: int = None,
    *args,
    **kwargs,
) -> pd.DataFrame:
    """
    Read a .smi (or .csv file with tab-separated values) in a pd.DataFrame.

    Args:
        filepath (str): path to a .smi file.
        chunk_size (int): size of the chunk. Defaults to None, a.k.a. no
            chunking.
        index_col (int): Data column used for indexing, defaults to 1.
        names (Sequence[str]): User-assigned names given to the columns.
        header (int): Row number to use as column names. Defaults to None.
        args (): Optional arguments for `pd.read_csv`.
        kwargs (): Optional keyword arguments for `pd.read_csv`.

    Returns:
        pd.DataFrame: a pd.DataFrame containing the data of the .smi file
            where the index is the index_col column.
    """
    try:
        return pd.read_csv(
            filepath,
            sep='\t',
            header=header,
            index_col=index_col,
            names=names,
            chunksize=chunk_size,
            *args,
            **kwargs,
        )
    except IndexError:
        raise IndexError(
            'Pandas does not understand the .smi file. The most common '
            'reason is a wrong delimiter (has to be \\t)'
        )



"""Protein language handling."""
from ..types import Tokenizer
from .processing import AA_FEAT, AA_PROPERTIES_NUM, BLOSUM62, BLOSUM62_NORM
from .protein_language import ProteinLanguage


class IndexesToSequenceError(Exception):
    pass


def token_indexes_to_sequence_raise(token_indexes: list) -> str:
    """monkey patch to raise Error."""
    raise IndexesToSequenceError(
        'token_indexes_to_sequence not implemented for '
        'binary_features since mapping is not unique.'
    )


class ProteinFeatureLanguage(ProteinLanguage):
    """
    ProteinFeatureLanguage class.

    ProteinFeatureLanguage handles Protein data and translates from text to
    feature space
    """

    def __init__(
        self,
        name: str = 'protein-feature-language',
        features: str = 'blosum',
        tokenizer: Tokenizer = list,
        add_start_and_stop: bool = True,
    ) -> None:
        """
        Initialize Protein feature language.

        Args:
            name (str): name of the ProteinFeatureLanguage.
            features (str): Feature alphabet choice. Defaults to 'blosum',
                alternatives are 'binary_features', 'float_features' and 'blosum_norm'.
            tokenizer (Tokenizer): This needs to be a function used to tokenize
                the amino acid sequences. The default is list which simply
                splits the sequence character-by-character.
        """
        self.name = name
        self.feat = features
        self.add_start_and_stop = add_start_and_stop

        if self.feat == 'binary_features':
            self.token_to_index = AA_PROPERTIES_NUM
            # monkey patching method
            self.token_indexes_to_sequence = token_indexes_to_sequence_raise
        elif self.feat == 'float_features':
            self.token_to_index = AA_FEAT
        elif self.feat == 'blosum':
            self.token_to_index = BLOSUM62
        elif self.feat == 'blosum_norm':
            self.token_to_index = BLOSUM62_NORM
        else:
            raise ValueError(
                "Choose dict as 'binary_features', 'float_features', 'blosum' or "
                f"'blosum_norm' (given was {features})."
            )

        self.number_of_features = len(self.token_to_index['<START>'])
        self.tokenizer = tokenizer

        self.setup_dict()

    def token_indexes_to_sequence(self, token_indexes: list) -> str:
        """
        Transform a list of tuples of token indexes into amino acid sequence.

        Args:
            token_indexes (list): a list of tuples, one tuple per AA and each
                tuple has length self.number_of_features

        Returns:
            str: an amino acid sequence representation.
        """
        return ''.join(
            [
                self.index_to_token.get(token_index, '')
                for token_index in token_indexes
                if token_index in self.sequence_tokens
            ]
        )

    @property
    def method(self) -> str:
        """A string denoting the language encoding method"""
        return self.feat



from rdkit import Chem

FORBIDDEN = set(['B', 'O', 'U', 'X', 'Z'])


def aas_to_smiles(aas, sanitize=True):
    """Converts an amino acid sequence (IUPAC) into SMILES.

    Args:
        aas (str): The amino acid sequence to be converted.
            Following IUPAC notation.
        sanitize (bool, optional): [description]. Defaults to True.

    Raises:
        TypeError: If aas is not a string.
        ValueError: If string cannot be converted to mol.

    Returns:
        smiles: SMILES string of the AA sequence.
    """
    if not isinstance(aas, str):
        raise TypeError(f'Provide string not {type(aas)}.')
    if len(set(aas).intersection(FORBIDDEN)) > 0:
        raise ValueError(
            f'Characters from: {FORBIDDEN} cant be parsed. Found one in: {aas}'
        )
    mol = Chem.MolFromFASTA(aas, sanitize=sanitize)
    if mol is None:
        raise ValueError(f'Sequence could not be converted to SMILES: {aas}')
    smiles = Chem.MolToSmiles(mol)
    return smiles



"""Polymer language handling."""
from typing import Sequence

from ..types import Indexes, Tensor, Union  # , delegate_kwargs
from .smiles_language import SMILESTokenizer
from .transforms import compose_encoding_transforms, compose_smiles_transforms


# @delegate_kwargs
class PolymerTokenizer(SMILESTokenizer):
    """
    PolymerTokenizer class.

    PolymerTokenizer is an extension of SMILESTokenizer adding special start and
    stop tokens per entity.
    A polymer language is usually shared across several SMILES datasets (e.g.
    different entity sources).
    """

    def __init__(
        self,
        entity_names: Sequence[str],
        name: str = 'polymer-language',
        add_start_and_stop: bool = True,
        **kwargs,
    ) -> None:
        """
        Initialize Polymer language able to encode different entities.

        Args:
            entity_names (Sequence[str]): A list of entity names that the
                polymer language can distinguish.
            name (str): name of the PolymerTokenizer.
            add_start_and_stop (bool): add start and stop token indexes.
                Defaults to True.
            kwargs (dict): additional parameters passed to SMILESTokenizer.

        NOTE:
            See `set_smiles_transforms` and `set_encoding_transforms` to change
            the transforms temporarily and reset with
            `reset_initial_transforms`. Assignment of class attributes
            in the parameter list will trigger such a reset.
        """

        super().__init__(name=name, add_start_and_stop=add_start_and_stop, **kwargs)
        self.entities = list(map(lambda x: x.capitalize(), entity_names))
        self.init_kwargs['entity_names'] = self.entities
        self.current_entity = None

        # rebuild basic vocab to group special tokens
        self.start_entity_tokens, self.stop_entity_tokens = (
            list(map(lambda x: '<' + x.upper() + '_' + s + '>', entity_names))
            for s in ['START', 'STOP']
        )

        # required for `token_indexes_to_smiles`
        self.special_indexes.update(
            enumerate(
                self.start_entity_tokens + self.stop_entity_tokens,
                start=len(self.special_indexes),
            )
        )
        self.setup_vocab()

        if kwargs.get('vocab_file', None):
            self.load_vocabulary(kwargs['vocab_file'])

        self.reset_initial_transforms()

    def _check_entity(self, entity: str) -> str:
        entity_ = entity.capitalize()
        if entity_ not in self.entities:
            raise ValueError(f'Unknown entity was given ({entity_})')
        return entity_

    def update_entity(self, entity: str) -> None:
        """
        Update the current entity and the default transforms (used e.g. in
        `add_dataset`) of the Polymer language object.

        Args:
            entity (str): a chemical entity (e.g. 'Monomer').
        """

        self.current_entity = self._check_entity(entity)
        self.transform_smiles = self.all_smiles_transforms[self.current_entity]
        self.transform_encoding = self.all_encoding_transforms[self.current_entity]

    def smiles_to_token_indexes(
        self, smiles: str, entity: str = None
    ) -> Union[Indexes, Tensor]:
        """
        Transform character-level SMILES into a sequence of token indexes.

        In case of add_start_stop, inserts entity specific tokens.

        Args:
            smiles (str): a SMILES (or SELFIES) representation.
            entity (str): a chemical entity (e.g. 'Monomer'). Defaults to
                None, where the current entity is used (initially the
                SMILESTokenizer default).

        Returns:
            Union[Indexes, Tensor]: indexes representation for the
                SMILES/SELFIES provided.
        """
        if entity is None:
            # default behavior given by call to update_entity()
            entity = self.current_entity
        else:
            entity = self._check_entity(entity)

        return self.all_encoding_transforms[entity](
            [
                self.token_to_index.get(token, self.unknown_token)
                for token in self.smiles_tokenizer(
                    self.all_smiles_transforms[entity](smiles)
                )
            ]
        )

    def reset_initial_transforms(self):
        """
        Reset smiles and token indexes transforms as on initialization,
        including entity specific transforms.
        """
        super().reset_initial_transforms()
        if not hasattr(self, 'entities'):  # call from base
            return
        self.current_entity = None
        self.all_smiles_transforms = {
            None: self.transform_smiles,
        }
        self.all_encoding_transforms = {
            None: self.transform_encoding,
        }
        for entity in self.entities:
            self.set_smiles_transforms(entity)
            self.set_encoding_transforms(entity)

    def set_smiles_transforms(
        self,
        entity,
        canonical=None,
        augment=None,
        kekulize=None,
        all_bonds_explicit=None,
        all_hs_explicit=None,
        remove_bonddir=None,
        remove_chirality=None,
        selfies=None,
        sanitize=None,
    ):
        """
        Helper function to reversibly change the transforms per entity.
        """
        entity = self._check_entity(entity)
        self.all_smiles_transforms[entity] = compose_smiles_transforms(
            canonical=canonical if canonical is not None else self.canonical,
            augment=augment if augment is not None else self.augment,
            kekulize=kekulize if kekulize is not None else self.kekulize,
            all_bonds_explicit=all_bonds_explicit
            if all_bonds_explicit is not None
            else self.all_bonds_explicit,
            all_hs_explicit=all_hs_explicit
            if all_hs_explicit is not None
            else self.all_hs_explicit,
            remove_bonddir=remove_bonddir
            if remove_bonddir is not None
            else self.remove_bonddir,
            remove_chirality=remove_chirality
            if remove_chirality is not None
            else self.remove_chirality,
            selfies=selfies if selfies is not None else self.selfies,
            sanitize=sanitize if sanitize is not None else self.sanitize,
        )

    def set_encoding_transforms(
        self,
        entity,
        randomize=None,
        add_start_and_stop=None,
        padding=None,
        padding_length=None,
        device=None,
    ):
        """
        Helper function to reversibly change the transforms per entity.
        Addresses entity specific start and stop tokens.
        """
        entity = self._check_entity(entity)
        start_index = self.token_to_index['<' + entity.upper() + '_START>']
        stop_index = self.token_to_index['<' + entity.upper() + '_STOP>']

        self.all_encoding_transforms[entity] = compose_encoding_transforms(
            randomize=randomize if randomize is not None else self.randomize,
            add_start_and_stop=add_start_and_stop
            if add_start_and_stop is not None
            else self.add_start_and_stop,
            start_index=start_index,
            stop_index=stop_index,
            padding=padding if padding is not None else self.padding,
            padding_length=padding_length
            if padding_length is not None
            else self.padding_length,
            padding_index=self.padding_index,
            device=device if device is not None else self.device,
        )
        if add_start_and_stop is not None:
            self._set_token_len_fn(add_start_and_stop)



"""Testing DrugSensitivityDataset."""
import os
import unittest

import numpy as np
from torch.utils.data import DataLoader

from pytoda.datasets import DrugSensitivityDataset
from pytoda.tests.utils import TestFileContent

COLUMN_NAMES = [',drug,cell_line,IC50', ',molecule,omic,label']
DRUG_SENSITIVITY_CONTENT = os.linesep.join(
    [
        '0,CHEMBL14688,sample_3,2.1',
        '1,CHEMBL14688,sample_2,-0.9',
        '2,CHEMBL17564,sample_1,1.2',
        '3,CHEMBL17564,sample_2,1.5',
    ]
)
SMILES_CONTENT = os.linesep.join(
    ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
)
GENE_EXPRESSION_CONTENT = os.linesep.join(
    [
        'genes,A,B,C,D',
        'sample_3,9.45,4.984,7.016,8.336',
        'sample_2,7.188,0.695,10.34,6.047',
        'sample_1,9.25,6.133,5.047,5.6',
    ]
)


class TestDrugSensitivityDatasetEagerBackend(unittest.TestCase):
    """Testing DrugSensitivityDataset with eager backend."""

    def setUp(self):
        self.backend = 'eager'
        print(f'backend is {self.backend}')
        self.smiles_content = SMILES_CONTENT
        self.gene_expression_content = GENE_EXPRESSION_CONTENT

        for column_names in COLUMN_NAMES:
            self.drug_sensitivity_content = os.linesep.join(
                [column_names, DRUG_SENSITIVITY_CONTENT]
            )

            with TestFileContent(
                self.drug_sensitivity_content
            ) as drug_sensitivity_file:
                with TestFileContent(self.smiles_content) as smiles_file:
                    with TestFileContent(
                        self.gene_expression_content
                    ) as gene_expression_file:
                        self.drug_sensitivity_dataset = DrugSensitivityDataset(
                            drug_sensitivity_file.filename,
                            smiles_file.filename,
                            gene_expression_file.filename,
                            gene_expression_kwargs={'pandas_dtype': {'genes': str}},
                            backend=self.backend,
                            column_names=column_names.split(',')[1:],
                        )

    def test___len__(self) -> None:
        """Test __len__."""
        self.assertEqual(len(self.drug_sensitivity_dataset), 4)

    def test___getitem__(self) -> None:
        """Test __getitem__."""

        padding_index = (
            self.drug_sensitivity_dataset.smiles_dataset.smiles_language.padding_index
        )
        c_index = (
            self.drug_sensitivity_dataset.smiles_dataset.smiles_language.token_to_index[
                'C'
            ]
        )
        o_index = (
            self.drug_sensitivity_dataset.smiles_dataset.smiles_language.token_to_index[
                'O'
            ]
        )
        (
            token_indexes_tensor,
            gene_expression_tensor,
            ic50_tensor,
        ) = self.drug_sensitivity_dataset[0]
        np.testing.assert_almost_equal(
            token_indexes_tensor.numpy(),
            np.array([padding_index, padding_index, c_index, o_index]),
        )
        np.testing.assert_almost_equal(
            gene_expression_tensor.numpy(),
            self.drug_sensitivity_dataset.gene_expression_dataset.get_item_from_key(
                'sample_3'
            ).numpy(),
        )
        np.testing.assert_almost_equal(ic50_tensor.numpy(), np.array([1.0]))

    def test_data_loader(self) -> None:
        """Test data_loader."""

        data_loader = DataLoader(
            self.drug_sensitivity_dataset, batch_size=2, shuffle=True
        )
        for (
            batch_index,
            (token_indexes_batch, gene_expression_batch, ic50_batch),
        ) in enumerate(data_loader):
            self.assertEqual(token_indexes_batch.size(), (2, 4))
            self.assertEqual(gene_expression_batch.size(), (2, 4))
            self.assertEqual(ic50_batch.size(), (2, 1))
            if batch_index > 4:
                break


class TestDrugSensitivityDatasetLazyBackend(
    TestDrugSensitivityDatasetEagerBackend
):  # noqa
    """Testing DrugSensitivityDataset with lazy backend."""

    def setUp(self):
        self.backend = 'lazy'
        print(f'backend is {self.backend}')
        self.smiles_content = SMILES_CONTENT
        self.gene_expression_content = GENE_EXPRESSION_CONTENT

        for column_names in COLUMN_NAMES:
            self.drug_sensitivity_content = os.linesep.join(
                [column_names, DRUG_SENSITIVITY_CONTENT]
            )
            with TestFileContent(
                self.drug_sensitivity_content
            ) as drug_sensitivity_file:
                with TestFileContent(self.smiles_content) as smiles_file:
                    with TestFileContent(
                        self.gene_expression_content
                    ) as gene_expression_file:
                        self.drug_sensitivity_dataset = DrugSensitivityDataset(
                            drug_sensitivity_file.filename,
                            smiles_file.filename,
                            gene_expression_file.filename,
                            gene_expression_kwargs={'pandas_dtype': {'genes': str}},
                            backend=self.backend,
                            column_names=column_names.split(',')[1:],
                        )


if __name__ == '__main__':
    unittest.main()



"""Amino Acid Sequence transforms."""
from ..transforms import Transform
from ..types import Indexes
from .protein_language import ProteinLanguage


class SequenceToTokenIndexes(Transform):
    """Transform Sequence to token indexes using Sequence language."""

    def __init__(self, protein_language: ProteinLanguage) -> None:
        """
        Initialize a Sequence to token indexes object.

        Args:
            protein_language (ProteinLanguage): a Protein language.
        """
        self.protein_language = protein_language

    def __call__(self, smiles: str) -> Indexes:
        """
        Apply the Sequence tokenization transformation

        Args:
            smiles (str): a Sequence representation.

        Returns:
            Indexes: indexes representation for the Sequence provided.
        """
        return self.protein_language.sequence_to_token_indexes(smiles)



"""Testing AnnotatedDataset dataset with eager backend."""
import os
import unittest

import numpy as np

from pytoda.datasets import AnnotatedDataset, SMILESTokenizerDataset, indexed, keyed
from pytoda.tests.utils import TestFileContent

# must contain all keys in annotated
SMILES_CONTENT = os.linesep.join(
    ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
)
ANNOTATED_CONTENT = os.linesep.join(
    [
        'label_0,label_1,annotation_index',
        '2.3,3.4,CHEMBL545',
        '4.5,5.6,CHEMBL17564',
        '6.7,7.8,CHEMBL602',
    ]
)


class TestAnnotatedDataset(unittest.TestCase):
    """Testing annotated dataset."""

    def setUp(self):
        self.smiles_content = SMILES_CONTENT
        self.annotated_content = ANNOTATED_CONTENT

    def test___getitem__(self) -> None:
        """Test __getitem__."""
        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = SMILESTokenizerDataset(
                    smiles_file.filename, add_start_and_stop=True, backend='eager'
                )
                annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=smiles_dataset
                )
        pad_index = smiles_dataset.smiles_language.padding_index
        start_index = smiles_dataset.smiles_language.start_index
        stop_index = smiles_dataset.smiles_language.stop_index
        c_index = smiles_dataset.smiles_language.token_to_index['C']
        o_index = smiles_dataset.smiles_language.token_to_index['O']
        n_index = smiles_dataset.smiles_language.token_to_index['N']
        s_index = smiles_dataset.smiles_language.token_to_index['S']
        # test first sample
        smiles_tokens, labels = annotated_dataset[0]
        self.assertEqual(
            smiles_tokens.numpy().flatten().tolist(),
            [pad_index, start_index, c_index, c_index, o_index, stop_index],
        )
        self.assertTrue(np.allclose(labels.numpy().flatten().tolist(), [2.3, 3.4]))
        # test last sample
        smiles_tokens, labels = annotated_dataset[2]
        self.assertEqual(
            smiles_tokens.numpy().flatten().tolist(),
            [start_index, n_index, c_index, c_index, s_index, stop_index],
        )
        self.assertTrue(np.allclose(labels.numpy().flatten().tolist(), [6.7, 7.8]))

    def test___getitem___from_indexed_annotation(self) -> None:
        """Test __getitem__ with index in the annotation file."""
        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = SMILESTokenizerDataset(
                    smiles_file.filename, add_start_and_stop=True, backend='eager'
                )
                annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=smiles_dataset
                )

        pad_index = smiles_dataset.smiles_language.padding_index
        start_index = smiles_dataset.smiles_language.start_index
        stop_index = smiles_dataset.smiles_language.stop_index
        c_index = smiles_dataset.smiles_language.token_to_index['C']
        o_index = smiles_dataset.smiles_language.token_to_index['O']
        n_index = smiles_dataset.smiles_language.token_to_index['N']
        s_index = smiles_dataset.smiles_language.token_to_index['S']
        # test first sample
        smiles_tokens, labels = annotated_dataset[0]
        self.assertEqual(
            smiles_tokens.numpy().flatten().tolist(),
            [pad_index, start_index, c_index, c_index, o_index, stop_index],
        )
        self.assertTrue(np.allclose(labels.numpy().flatten().tolist(), [2.3, 3.4]))
        # test last sample
        smiles_tokens, labels = annotated_dataset[2]
        self.assertEqual(
            smiles_tokens.numpy().flatten().tolist(),
            [start_index, n_index, c_index, c_index, s_index, stop_index],
        )
        self.assertTrue(np.allclose(labels.numpy().flatten().tolist(), [6.7, 7.8]))

    def _test_indexed(self, ds, keys, index):
        key = keys[index]
        positive_index = index % len(ds)
        # get_key (support for negative index?)
        self.assertEqual(key, ds.get_key(positive_index))
        self.assertEqual(key, ds.get_key(index))
        # get_index
        self.assertEqual(positive_index, ds.get_index(key))
        # get_item_from_key returning labels as well
        for from_index, from_key in zip(ds[index], ds.get_item_from_key(key)):
            self.assertTrue(all(from_index == from_key))
        # keys
        self.assertSequenceEqual(keys, list(ds.keys()))
        # duplicate keys
        self.assertFalse(ds.has_duplicate_keys)

    def test_all_base_for_indexed_methods(self):

        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = SMILESTokenizerDataset(
                    smiles_file.filename, add_start_and_stop=True, backend='eager'
                )
                annotated_dataset = AnnotatedDataset(
                    annotation_file.filename,
                    dataset=smiles_dataset,
                    index_col=0,
                    label_columns=['label_1'],
                )
                duplicate_ds = AnnotatedDataset(
                    annotation_file.filename,
                    dataset=smiles_dataset + smiles_dataset,
                )

        all_keys = [
            row.split(',')[-1] for row in self.annotated_content.split(os.linesep)[1:]
        ]

        for ds, keys in [
            (annotated_dataset, all_keys),
        ]:
            index = -1
            self._test_indexed(ds, keys, index)

        # duplicates in datasource can be checked directly
        self.assertTrue(duplicate_ds.datasource.has_duplicate_keys)
        # DataFrame is the dataset
        self.assertFalse(duplicate_ds.has_duplicate_keys)


class TestChangeIndexingReturn(unittest.TestCase):
    """Testing annotated dataset."""

    def setUp(self):
        self.smiles_content = SMILES_CONTENT
        self.annotated_content = ANNOTATED_CONTENT

    def test_return_integer_index(self) -> None:
        """Test __getitem__ with index in dataset."""
        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = SMILESTokenizerDataset(smiles_file.filename)
                # default
                default_annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=smiles_dataset
                )
                # outer
                indexed_annotated_dataset = indexed(default_annotated_dataset)
                # inner
                annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=indexed(smiles_dataset)
                )

        # default
        # relevant to check that `smiles_dataset` was not mutated from
        # the `indexed(smiles_dataset)` call
        smiles_tokens, labels = default_annotated_dataset[2]
        self.check_CHEMBL602(smiles_tokens, labels)

        # outer __getitem__
        (smiles_tokens, labels), sample_index = indexed_annotated_dataset[2]
        self.assertEqual(sample_index, 2)
        self.check_CHEMBL602(smiles_tokens, labels)
        # outer get_item_from_key
        (
            (smiles_tokens, labels),
            sample_index,
        ) = indexed_annotated_dataset.get_item_from_key('CHEMBL602')
        self.assertEqual(sample_index, 2)
        self.check_CHEMBL602(smiles_tokens, labels)

        # inner __getitem__
        (smiles_tokens, sample_index), labels = annotated_dataset[0]
        self.assertEqual(sample_index, 0)
        # inner __getitem__ with different index in smiles_dataset
        (smiles_tokens, sample_index), labels = annotated_dataset[2]
        self.assertEqual(sample_index, 3)
        self.check_CHEMBL602(smiles_tokens, labels)
        # inner get_item_from_key
        (smiles_tokens, sample_index), labels = annotated_dataset.get_item_from_key(
            'CHEMBL602'
        )
        self.assertEqual(sample_index, 3)
        self.check_CHEMBL602(smiles_tokens, labels)

    def test_return_key(self) -> None:
        """Test __getitem__ with key in dataset."""
        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = SMILESTokenizerDataset(smiles_file.filename)
                # default
                default_annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=smiles_dataset
                )
                # outer
                keyed_annotated_dataset = keyed(default_annotated_dataset)
                # inner
                annotated_dataset = AnnotatedDataset(
                    annotation_file.filename, dataset=keyed(smiles_dataset)
                )

        # default
        # relevant to check that `smiles_dataset` was not mutated from
        # the `keyed(smiles_dataset)` call
        smiles_tokens, labels = default_annotated_dataset[2]
        self.check_CHEMBL602(smiles_tokens, labels)

        # outer __getitem__
        (smiles_tokens, labels), sample_key = keyed_annotated_dataset[2]
        self.assertEqual(sample_key, 'CHEMBL602')
        self.check_CHEMBL602(smiles_tokens, labels)
        # outer get_item_from_key
        (smiles_tokens, labels), sample_key = keyed_annotated_dataset.get_item_from_key(
            'CHEMBL602'
        )
        self.assertEqual(sample_key, 'CHEMBL602')
        self.check_CHEMBL602(smiles_tokens, labels)

        # inner __getitem__
        (smiles_tokens, sample_key), labels = annotated_dataset[0]
        self.assertEqual(sample_key, 'CHEMBL545')
        # inner __getitem__ with different index in smiles_dataset
        (smiles_tokens, sample_key), labels = annotated_dataset[2]
        self.assertEqual(sample_key, 'CHEMBL602')
        self.check_CHEMBL602(smiles_tokens, labels)
        # inner get_item_from_key
        (smiles_tokens, sample_key), labels = annotated_dataset.get_item_from_key(
            'CHEMBL602'
        )
        self.assertEqual(sample_key, 'CHEMBL602')
        self.check_CHEMBL602(smiles_tokens, labels)

    def test_return_key_index_stacked(self) -> None:
        """Test __getitem__ with key in dataset."""
        with TestFileContent(self.smiles_content) as smiles_file:
            with TestFileContent(self.annotated_content) as annotation_file:
                smiles_dataset = keyed(
                    indexed(
                        SMILESTokenizerDataset(
                            smiles_file.filename,
                        )
                    )
                )
                annotated_dataset = indexed(
                    keyed(
                        AnnotatedDataset(
                            annotation_file.filename,
                            dataset=smiles_dataset,
                        )
                    )
                )

        (smiles_tokens, smiles_index), smiles_key = smiles_dataset[3]
        self.assertEqual(smiles_key, 'CHEMBL602')
        self.assertEqual(smiles_index, 3)
        self.check_CHEMBL602(smiles_tokens)
        (smiles_tokens, smiles_index), smiles_key = smiles_dataset.get_item_from_key(
            'CHEMBL602'
        )
        self.assertEqual(smiles_key, 'CHEMBL602')
        self.assertEqual(smiles_index, 3)
        self.check_CHEMBL602(smiles_tokens)

        (
            (
                (((smiles_tokens, smiles_index), smiles_key), labels),  # inner
                annotation_key,
            ),
            annotation_index,
        ) = annotated_dataset[2]
        self.assertEqual(smiles_key, 'CHEMBL602')
        self.assertEqual(smiles_index, 3)
        self.assertEqual(annotation_index, 2)
        self.assertEqual(annotation_key, 'CHEMBL602')
        (
            (
                (((smiles_tokens, smiles_index), smiles_key), labels),  # inner
                annotation_key,
            ),
            annotation_index,
        ) = annotated_dataset.get_item_from_key('CHEMBL602')
        self.assertEqual(smiles_key, 'CHEMBL602')
        self.assertEqual(smiles_index, 3)
        self.assertEqual(annotation_index, 2)
        self.assertEqual(annotation_key, 'CHEMBL602')

    def check_CHEMBL602(self, smiles_tokens, labels=None):
        """Check indexing results lack unwanted index/key from hidden calls"""
        self.assertEqual(len(smiles_tokens.numpy().flatten().tolist()), 4)
        if labels is not None:
            self.assertEqual(len(labels.numpy().flatten().tolist()), 2)


if __name__ == '__main__':
    unittest.main()



"""_CacheDatasource abstract implementation."""
import os
import shutil
import tempfile
import warnings

import diskcache as dc

from .utils import sizeof_fmt


class _CacheDatasource:
    """
    Supporting Datasets with underlying disk cache.

    Suggested when handling datasets that can't fit in the device memory.
    """

    size_limit = 1073741824  # default limit of 1GiB from diskcash

    def __init__(self, size_limit=None, fit_size_limit_filepath=None) -> None:
        """Initialize a dataset using a disk cache with maximal size."""
        if size_limit:
            self.size_limit = size_limit
        elif fit_size_limit_filepath:
            file_size = os.path.getsize(fit_size_limit_filepath)
            if file_size > self.size_limit:
                self.size_limit = file_size
                message = (
                    'Temporary directory for caching can be up to '
                    f'{self.size_limit} bytes ({sizeof_fmt(size_limit)}) '
                    'large to fit data.'
                )
                # ResourceWarning is usually filtered by default
                warnings.warn(message, ResourceWarning)
        self.cache_filepath = tempfile.mkdtemp()
        self.cache = dc.Cache(self.cache_filepath, size_limit=self.size_limit)

    def __del__(self):
        """Delete the _CacheDatasource."""
        self.cache.close()
        shutil.rmtree(self.cache_filepath)



"""Testing files processing."""
import os
import unittest

from pytoda.files import count_file_lines, read_smi
from pytoda.tests.utils import TestFileContent


class TestFiles(unittest.TestCase):
    """Testing files."""

    def test_count_file_lines(self) -> None:
        """Test count_file_lines."""
        content = os.linesep.join(['one batch', 'two batch', 'penny and dime'])
        with TestFileContent(content) as test_file:
            self.assertEqual(count_file_lines(test_file.filename), 3)
        content = os.linesep.join(
            ['one batch', 'two batch', 'penny and dime{}'.format(os.linesep)]
        )
        with TestFileContent(content) as test_file:
            self.assertEqual(count_file_lines(test_file.filename), 3)

    def test_read_smi(self) -> None:
        """Test read_smi."""
        content = os.linesep.join(
            [
                'CN(C)CCNC(=O)c1cc2CSc3cc(Cl)ccc3-c2s1\tdrug_0',
                'CC(C)N1C(=O)S\\C(=C\\c2ccc(Sc3nc4ccccc4[nH]3)o2)C1=O\tdrug_1',
                'C(Cn1c2ccccc2c2ccccc12)c1nc2ccccc2[nH]1\tdrug_2',
                'C1CN(CCO1)c1nnc(-c2ccccc2)c(n1)-c1ccccc1\tdrug_3',
            ]
        )
        indexes = ['drug_0', 'drug_1', 'drug_2', 'drug_3']
        smiles = [
            'CN(C)CCNC(=O)c1cc2CSc3cc(Cl)ccc3-c2s1',
            'CC(C)N1C(=O)S\\C(=C\\c2ccc(Sc3nc4ccccc4[nH]3)o2)C1=O',
            'C(Cn1c2ccccc2c2ccccc12)c1nc2ccccc2[nH]1',
            'C1CN(CCO1)c1nnc(-c2ccccc2)c(n1)-c1ccccc1',
        ]
        with TestFileContent(content) as test_file:
            smiles_df = read_smi(test_file.filename)
            self.assertEqual(smiles_df.shape, (4, 1))
            self.assertEqual(smiles_df.index.tolist(), indexes)
            for index, a_smiles in zip(indexes, smiles):
                self.assertEqual(smiles_df.loc[index]['SMILES'], a_smiles)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 Matteo Manica, Jannis Born, Ali Oskooei, Joris Cadow
# Most parts of this file are Licenced under the MIT Licence.
# Specifically the functions from_pretrained and save_pretrained are derivative
# works with sources under the following licence:
# Copyright 2018 The Open AI Team Authors and The HuggingFace Inc. team.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use these functions except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SMILES language handling."""
import copy
import json
import logging
import os
import warnings
from collections import Counter

import rdkit  # Needs import before torch in some envs
import dill
import torch
from selfies import decoder as selfies_decoder
from selfies import encoder as selfies_encoder

from ..files import read_smi
from ..transforms import Compose
from ..types import (
    Files,
    Indexes,
    Iterable,
    Sequence,
    Tensor,
    Tokenizer,
    Tokens,
    Tuple,
    Union,
)
from .processing import TOKENIZER_FUNCTIONS, tokenize_smiles
from .transforms import compose_encoding_transforms, compose_smiles_transforms

logger = logging.getLogger(__name__)

# mimicry of huggingface tokenizers
# see PreTrainedTokenizer
VOCAB_FILES_NAMES = {'vocab_file': 'vocab.json'}
# see PreTrainedTokenizerBase
TOKENIZER_CONFIG_FILE = 'tokenizer_config.json'
# our
TOKEN_COUNTS_FILE = 'token_count.json'


class UnknownMaxLengthError(RuntimeError):
    pass


class SMILESLanguage(object):
    """
    SMILESLanguage class.

    SMILESLanguage handle SMILES data defining the vocabulary and
    utilities to manipulate it, including encoding to token indexes.
    """

    vocab_files_names = VOCAB_FILES_NAMES

    def __init__(
        self,
        name: str = 'smiles-language',
        smiles_tokenizer: Tokenizer = tokenize_smiles,
        tokenizer_name: str = None,  # Literal only in python 3.8
        vocab_file: str = None,
        max_token_sequence_length: int = 0,
    ) -> None:
        """
        Initialize SMILES language.

        Args:
            name (str): name of the SMILESLanguage.
            smiles_tokenizer (Tokenizer): optional SMILES tokenization
                function. Defaults to tokenize_smiles, but tokenizer_name takes
                precedence when found in available TOKENIZER_FUNCTIONS.
            tokenizer_name (str): name, mapping to Tokenizer used to save and
                restore object from text files. Defaults to None, i.e.
                using default smiles_tokenizer. Examples for available names
                are 'smiles', 'selfies' or 'spe_smiles'.
            vocab_file (str): optional filepath to vocab json or directory
                containing it.
            max_token_sequence_length (int): initial value for keeping track
                of longest sequence. Defaults to 0.
        """
        self.name = name
        self.tokenizer_name = tokenizer_name
        if tokenizer_name is not None and tokenizer_name not in TOKENIZER_FUNCTIONS:
            logger.info(
                f'Given tokenizer_name {tokenizer_name} was not found, using'
                'default tokenizer function.'
            )
        self.smiles_tokenizer = TOKENIZER_FUNCTIONS.get(
            tokenizer_name, smiles_tokenizer
        )

        self.padding_token = '<PAD>'
        self.unknown_token = '<UNK>'
        self.start_token = '<START>'
        self.stop_token = '<STOP>'
        self.padding_index = 0
        self.unknown_index = 1
        self.start_index = 2
        self.stop_index = 3
        self.token_count = Counter()
        self.special_indexes = {
            self.padding_index: self.padding_token,
            self.unknown_index: self.unknown_token,
            self.start_index: self.start_token,
            self.stop_index: self.stop_token,
        }
        self.setup_vocab()

        if vocab_file:
            self.load_vocabulary(vocab_file)

        # updated when adding smiles
        self.max_token_sequence_length = max_token_sequence_length
        # updated by transformations, e.g. padding
        self._get_total_number_of_tokens_fn = len

        # inputs and kwargs for saving and re-loading (TOKENIZER_CONFIG_FILE)
        # (see ``from_pretrained`` and ``save_pretrained``)
        self.init_inputs = ()
        self.init_kwargs = {
            'name': self.name,
            'tokenizer_name': self.tokenizer_name,
            'max_token_sequence_length': self.max_token_sequence_length,
        }

        self.transform_smiles = Compose([])  # identity
        self.transform_encoding = Compose([])

    def setup_vocab(self) -> None:
        """
        Sets up the vocab by generating the special tokens.
        """
        # NOTE: include augmentation characters, parenthesis and numbers for
        #    rings
        additional_indexes_to_token = dict(
            enumerate(
                list('()')
                + list(map(str, range(1, 10)))
                + list('%{}'.format(index) for index in range(10, 30)),
                start=len(self.special_indexes),
            )
        )
        self.index_to_token = {**self.special_indexes, **additional_indexes_to_token}
        self.number_of_tokens = len(self.index_to_token)
        self.token_to_index = {
            token: index for index, token in self.index_to_token.items()
        }

    @staticmethod
    def load(filepath: str) -> 'SMILESLanguage':
        """
        Static method to load a SMILESLanguage object.

        Args:
            filepath (str): path to the file.

        Returns:
            SMILESLanguage: the loaded SMILES language object.
        """
        warnings.warn(
            "Loading languages will use a text files in the future", FutureWarning
        )
        with open(filepath, 'rb') as f:
            smiles_language = dill.load(f)
        return smiles_language

    @staticmethod
    def dump(smiles_language: 'SMILESLanguage', filepath: str):
        """
        Static method to save a smiles_language object to disk.

        Args:
            smiles_language (SMILESLanguage): a SMILESLanguage object.
            filepath (str): path where to dump the SMILESLanguage.
        """
        with open(filepath, 'wb') as f:
            dill.dump(smiles_language, f)

    def save(self, filepath: str):
        """
        Instance method to save/dump smiles language object.

        Args:
            filepath (str): path where to save the SMILESLanguage.
        """
        warnings.warn(
            "Saving languages will only store a text files in the future", FutureWarning
        )
        SMILESLanguage.dump(self, filepath)

    def load_vocabulary(self, vocab_file: str):
        """Load a vocabulary mapping from token to token indexes.

        Args:
            vocab_file (str): a .json with tokens mapping to index. Can also
                be path to directory.
        """
        if os.path.isdir(vocab_file):
            vocab_file = os.path.join(vocab_file, self.vocab_files_names['vocab_file'])

        with open(vocab_file, encoding="utf-8") as fp:
            vocab = json.load(fp)
        # encoder
        self.token_to_index = self._check_specials(vocab)
        # decoder
        self.index_to_token = {v: k for k, v in self.token_to_index.items()}
        self.number_of_tokens = len(self.index_to_token)

    def _check_specials(self, vocab):
        """Check that defined special tokens match class definitions."""
        for index, token in self.special_indexes.items():
            try:
                if vocab[token] != index:
                    warnings.warn(
                        f'The vocab does not have matching special tokens: '
                        f'{token} is {vocab[token]}, but was defined as '
                        f'{index}.',
                    )
            except KeyError:
                warnings.warn(f'The vocab is missing a special token: {token}.')
        return vocab

    @classmethod
    def from_pretrained(cls, pretrained_path, *init_inputs, **kwargs):
        # directory with vocab files
        # not handling ADDED_TOKENS_FILE or SPECIAL_TOKENS_MAP_FILE
        # only handle case of files on disk here
        # but include handling optional counts
        resolved_vocab_files = {}

        additional_files_names = {
            'tokenizer_config_file': TOKENIZER_CONFIG_FILE,
            'token_count_file': TOKEN_COUNTS_FILE,
        }

        # Look for the tokenizer main vocabulary files
        # and the additional tokens files
        if os.path.isdir(pretrained_path):
            for file_id, file_name in {
                **cls.vocab_files_names,
                **additional_files_names,
            }.items():
                full_file_name = os.path.join(pretrained_path, file_name)
                if not os.path.exists(full_file_name):
                    logger.info(
                        "Didn't find file {}. We won't load it.".format(full_file_name)
                    )
                    full_file_name = None

                resolved_vocab_files[file_id] = full_file_name

        # Prepare tokenizer initialization kwargs
        tokenizer_config_file = resolved_vocab_files.pop('tokenizer_config_file', None)
        if tokenizer_config_file is not None:
            with open(tokenizer_config_file, encoding='utf-8') as config_file:
                init_kwargs = json.load(config_file)
            saved_init_inputs = init_kwargs.pop("init_inputs", ())
            if not init_inputs:
                init_inputs = saved_init_inputs
        else:
            init_kwargs = {}

        # Update with newly provided kwargs
        init_kwargs.update(kwargs)

        token_count_file = resolved_vocab_files.pop("token_count_file", None)

        # adds remaining (i.e. vocab_file) to kwargs
        for args_name, file_path in resolved_vocab_files.items():
            if args_name not in init_kwargs:
                init_kwargs[args_name] = file_path

        # Instantiate tokenizer.
        try:
            tokenizer = cls(*init_inputs, **init_kwargs)
        except OSError:
            raise OSError(
                'Unable to load vocabulary from file. '
                'Please check that the provided vocabulary is accessible '
                'and not corrupted.'
            )
        if token_count_file is not None:
            with open(token_count_file, encoding='utf-8') as counts_file:
                tokenizer.token_count = Counter(json.load(counts_file))

        # set args and kwargs explicitly here.
        tokenizer.init_inputs = init_inputs
        tokenizer.init_kwargs = init_kwargs

        return tokenizer

    def save_vocabulary(self, vocab_file: str) -> Tuple[str]:
        """Save the vocabulary mapping tokens to indexes to file.

        Args:
            vocab_file (str): a .json to save tokens mapping to index. Can also
                be path to directory.
        """
        if os.path.isdir(vocab_file):
            vocab_file = os.path.join(vocab_file, self.vocab_files_names['vocab_file'])

        with open(vocab_file, 'w', encoding="utf-8") as fp:
            json.dump(self.token_to_index, fp, indent=4)

        return (vocab_file,)

    def save_pretrained(self, save_directory):
        """Save the tokenizer vocabulary files together with
        tokenizer instantiation positional and keywords inputs.

        This method make sure the full tokenizer can then be re-loaded
        using the `from_pretrained` class method.
        """
        if not os.path.isdir(save_directory):
            # TODO raise?
            logger.error(
                'Saving directory ({}) should be a directory'.format(save_directory)
            )
            return

        tokenizer_config_file = os.path.join(save_directory, TOKENIZER_CONFIG_FILE)
        tokenizer_counts_file = os.path.join(save_directory, TOKEN_COUNTS_FILE)

        tokenizer_config = copy.deepcopy(self.init_kwargs)
        if len(self.init_inputs) > 0:
            tokenizer_config['init_inputs'] = copy.deepcopy(self.init_inputs)
        for file_id in self.vocab_files_names.keys():
            tokenizer_config.pop(file_id, None)

        with open(tokenizer_config_file, 'w', encoding='utf-8') as fp:
            json.dump(tokenizer_config, fp=fp, ensure_ascii=False, indent=4)

        with open(tokenizer_counts_file, 'w', encoding='utf-8') as fp:
            json.dump(self.token_count, fp=fp, ensure_ascii=False, indent=4)

        vocab_files = self.save_vocabulary(save_directory)

        return vocab_files + (tokenizer_counts_file,)

    def _load_vocabulary_from_pickled_language(
        self, filepath: str, include_metadata: bool = False
    ) -> None:
        """Save the vocabulary mapping tokens to indexes from file.

        Args:
            filepath (str): path to the dump of the SMILESLanguage.
        """
        a_language = self.load(filepath)
        # encoder
        self.token_to_index = self._check_specials(a_language.token_to_index)
        # decoder
        self.index_to_token = {v: k for k, v in self.token_to_index.items()}
        self.number_of_tokens = len(self.index_to_token)

    def _from_legacy_pickled_language(self, filepath: str) -> None:
        """Load a current language instance from pickled legacy language.

        Args:
            filepath (str): path to the dump of the SMILESLanguage.
        """
        warnings.warn(
            "Loading from legacy languages will be deprecated", DeprecationWarning
        )
        a_language = self.load(filepath)
        # encoder
        # missing special tokens
        self.token_to_index = a_language.token_to_index
        self.token_to_index.update({t: i for i, t in self.special_indexes.items()})
        # decoder
        self.index_to_token = {v: k for k, v in self.token_to_index.items()}
        self.number_of_tokens = len(self.index_to_token)

        self.max_token_sequence_length = a_language.max_token_sequence_length  # noqa
        self.init_kwargs['max_token_sequence_length'] = self.max_token_sequence_length
        self.token_count = a_language._token_count

    def _update_max_token_sequence_length(self, tokens: Tokens) -> None:
        """
        Update the max token sequence length.
        Uses method possibly overloaded by transformation setup to assess the
        length of tokens after transformations prior to their application.
        For example this allows handling start and stop tokens.

        Args:
            tokens (Tokens): tokens considered.
        """
        total_number_of_tokens = self._get_total_number_of_tokens_fn(tokens)
        if total_number_of_tokens > self.max_token_sequence_length:
            self.max_token_sequence_length = total_number_of_tokens
            self.init_kwargs['max_token_sequence_length'] = total_number_of_tokens

    def _update_language_dictionaries_with_tokens(self, tokens: Tokens) -> None:
        """
        Update the language dictionaries with provided tokens.

        Args:
            tokens (Tokens): tokens considered.
        """
        # count tokens
        tokens_counter = Counter(tokens)
        # index to token
        index_to_token = dict(
            enumerate(
                tokens_counter.keys() - self.token_to_index.keys(),
                self.number_of_tokens,
            )
        )
        # update language
        self.token_count += tokens_counter
        self.index_to_token.update(index_to_token)
        self.token_to_index.update(
            {token: index for index, token in index_to_token.items()}
        )
        self.number_of_tokens += len(index_to_token)

    def add_smis(
        self,
        smi_filepaths: Files,
        index_col: int = 1,
        chunk_size: int = 10000,
        name: str = 'SMILES',
        names: Sequence[str] = None,
    ) -> None:
        """
        Add a set of SMILES from a list of .smi files, applying
        `transform_smiles`.

        Args:
            smi_filepaths (Files): a list of paths to .smi files.
            index_col (int): Data column used for indexing, defaults to 1.
            chunk_size (int): size of the chunks. Defaults to 10000.
            name (str): type of dataset, used to index columns in smi, and must
                be in names. Defaults to 'SMILES'.
            names (Sequence[str]): User-assigned names given to the columns.
                Defaults to `[name]`.
        """
        for smi_filepath in smi_filepaths:
            self.add_smi(
                smi_filepath,
                index_col=index_col,
                chunk_size=chunk_size,
                name=name,
                names=names,
            )

    def add_smi(
        self,
        smi_filepath: str,
        index_col: int = 1,
        chunk_size: int = 10000,
        name: str = 'SMILES',
        names: Sequence[str] = None,
    ) -> None:
        """
        Add a set of SMILES from a .smi file, applying `transform_smiles`.

        Args:
            smi_filepath (str): path to the .smi file.
            index_col (int): Data column used for indexing, defaults to 1.
            chunk_size (int): number of rows to read in a chunk.
                Defaults to 100000.
            name (str): type of dataset, used to index columns in smi, and must
                be in names. Defaults to 'SMILES'.
            names (Sequence[str]): User-assigned names given to the columns.
                Defaults to `[name]`.
        """
        names = names or [name]
        try:
            for chunk in read_smi(
                smi_filepath, index_col=index_col, chunk_size=chunk_size, names=names
            ):
                for smiles in chunk[name]:
                    try:
                        transformed_smiles = self.transform_smiles(smiles)
                        self.add_smiles(transformed_smiles)
                    except Exception:
                        logger.warning(
                            'transformation of smiles or adding result to '
                            f'the language failed for: {smiles}'
                        )
        except IndexError:
            raise IndexError('There must be one name per column in names.')
        except KeyError as error:
            raise KeyError(
                f'{str(error)}. Check index_col and that name {name} is in '
                f' names {names}'
            )

    def add_dataset(self, dataset: Iterable):
        """
        Add a set of SMILES from an iterable, applying `transform_smiles`.

        Collects and warns about invalid SMILES, and warns on finding new
        tokens.

        Args:
            dataset (Iterable): returning SMILES strings.
        """
        initial_vocab_length = len(self.token_to_index)
        self.invalid_molecules = []
        self.failed_transform_smiles = []

        for index, smiles in enumerate(dataset):
            if rdkit.Chem.MolFromSmiles(smiles, sanitize=False) is None:
                self.invalid_molecules.append((index, smiles))
            else:
                try:
                    transformed_smiles = self.transform_smiles(smiles)
                except Exception:
                    self.failed_transform_smiles.append((index, smiles))
                else:
                    self.add_smiles(transformed_smiles)

        # Raise warning about invalid molecules
        if len(self.invalid_molecules) > 0:
            logger.warning(
                f'NOTE: We found {len(self.invalid_molecules)} invalid  '
                'smiles. Check the warning trace and inspect the  attribute '
                '`invalid_molecules`. To remove invalid  SMILES in your .smi '
                'file, we recommend using '
                '`pytoda.preprocessing.smi.smi_cleaner`.'
            )
        # Raise warning about failed transformations
        if len(self.failed_transform_smiles) > 0:
            logger.warning(
                f'NOTE: We found {len(self.failed_transform_smiles)} smiles '
                'that failed to be transformed (excluding invalid smiles). '
                'Inspect the attribute `failed_transform_smiles`.'
            )

        # Raise warning if new tokens were added.
        if len(self.token_to_index) > initial_vocab_length:
            logger.warning(
                f'{len(self.token_to_index) - initial_vocab_length}'
                ' new token(s) were added to SMILES language.'
            )

    def add_smiles(self, smiles: str) -> None:
        """
        Add a SMILES to the language.

        Updates `max_token_sequence_length`.
        Adds missing tokens to the language.

        Args:
            smiles (str): a SMILES representation.
        """
        tokens = self.smiles_tokenizer(smiles)
        self._update_max_token_sequence_length(tokens)
        self._update_language_dictionaries_with_tokens(tokens)

    def add_token(self, token: str) -> None:
        """
        Add a token to the language.

        Args:
            token (str): a token.
        """
        if token in self.token_to_index:
            self.token_count[token] += 1
        else:
            self.token_to_index[token] = self.number_of_tokens
            self.token_count[token] = 1
            self.index_to_token[self.number_of_tokens] = token
            self.number_of_tokens += 1

    def smiles_to_token_indexes(self, smiles: str) -> Union[Indexes, Tensor]:
        """
        Transform character-level SMILES into a sequence of token indexes.

        Args:
            smiles (str): a SMILES (or SELFIES) representation.

        Returns:
            Union[Indexes, Tensor]: indexes representation for the
                SMILES/SELFIES provided.
        """
        return self.transform_encoding(
            [
                self.token_to_index.get(token, self.unknown_index)
                for token in self.smiles_tokenizer(self.transform_smiles(smiles))
            ]
        )

    def token_indexes_to_smiles(self, token_indexes: Union[Indexes, Tensor]) -> str:
        """
        Transform a sequence of token indexes into SMILES, ignoring special
        tokens.

        Args:
            token_indexes (Union[Indexes, Tensor]): Sequence of integers
                representing tokens in vocabulary.

        Returns:
            str: a SMILES (or SELFIES) representation.
        """
        token_indexes = self.tensor_to_indexes(token_indexes)

        return ''.join(
            [
                self.index_to_token.get(token_index, '')
                for token_index in token_indexes
                # consider only valid SMILES token indexes
                if token_index not in self.special_indexes
            ]
        )

    @staticmethod
    def tensor_to_indexes(token_indexes: Union[Indexes, Tensor]) -> Indexes:
        """Utility to get Indexes from Tensors.

        Args:
            token_indexes (Union[Indexes, Tensor]): from single SMILES.

        Raises:
            ValueError: in case the Tensor is not shaped correctly

        Returns:
            Indexes: list from Tensor or else the initial token_indexes.
        """
        if isinstance(token_indexes, torch.Tensor):
            if token_indexes.ndim != 1:
                raise ValueError('Only token indexes for a single SMILES are supported')
            return token_indexes.numpy().flatten().tolist()

        return token_indexes

    def selfies_to_smiles(self, selfies: str) -> str:
        """
        SELFIES to SMILES converter method.
        Based on: https://arxiv.org/abs/1905.13741

        Arguments:
            selfies (str): SELFIES representation

        Returns:
            str: A SMILES string
        """
        if not isinstance(selfies, str):
            raise TypeError(f'Wrong data type: {type(selfies)}. Use strings.')
        try:
            return selfies_decoder(selfies)
        except Exception:
            logger.warning(
                f'Could not convert SELFIES {selfies} to SMILES, returning '
                'the SELFIES instead'
            )
            return selfies

    def smiles_to_selfies(self, smiles: str) -> str:
        """
        SMILES to SELFIES converter method.
        Based on: https://arxiv.org/abs/1905.13741

        Arguments:
            smiles (str): SMILES representation

        Returns:
            str: A SELFIES string
        """
        if not isinstance(smiles, str):
            raise TypeError(f'Wrong data type: {type(smiles)}. Use strings.')
        try:
            return selfies_encoder(smiles)
        except Exception:
            logger.warning(
                f'Could not convert SMILES {smiles} to SELFIES, returning '
                'the SMILES instead'
            )
            return smiles


class SELFIESLanguage(SMILESLanguage):
    """
    SELFIESLanguage is a SMILESLanguage with a different default tokenizer,
    transforming SMILES to SELFIES.
    """

    def __init__(
        self,
        name: str = 'selfies-language',
        vocab_file: str = None,
        max_token_sequence_length: int = 0,
    ) -> None:
        """
        Initialize SMILES language.

        Args:
            name (str): name of the SMILESLanguage.
            vocab_file (str): optional filepath to vocab json or directory
                containing it.
            max_token_sequence_length (int): initial value for keeping track
                of longest sequence. Defaults to 0.
        """
        super().__init__(
            name=name,
            tokenizer_name='selfies',
            vocab_file=vocab_file,
            max_token_sequence_length=max_token_sequence_length,
        )
        self.transform_smiles = selfies_encoder


class SMILESTokenizer(SMILESLanguage):
    """
    SMILESTokenizer class, based on SMILESLanguage applying transforms and
    and encoding of SMILES string to sequence of token indexes.
    """

    def __init__(
        self,
        name: str = 'smiles-language',
        smiles_tokenizer: Tokenizer = tokenize_smiles,
        tokenizer_name: str = None,
        vocab_file: str = None,
        max_token_sequence_length: int = 0,
        canonical: bool = False,
        augment: bool = False,
        kekulize: bool = False,
        all_bonds_explicit: bool = False,
        all_hs_explicit: bool = False,
        remove_bonddir: bool = False,
        remove_chirality: bool = False,
        selfies: bool = False,
        sanitize: bool = True,
        randomize: bool = False,
        add_start_and_stop: bool = False,
        padding: bool = False,
        padding_length: int = None,
        device: torch.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu'
        ),
    ) -> None:
        """
        Initialize SMILES language.

        Args:
            name (str): name of the SMILESLanguage.
            smiles_tokenizer (Tokenizer): optional SMILES tokenization
                function. Defaults to tokenize_smiles, but tokenizer_name takes
                precedence when found in available TOKENIZER_FUNCTIONS.
            tokenizer_name (str): optional name mapping to Tokenizer. Defaults
                to None, i.e. using default smiles_tokenizer.
            vocab_file (str): optional filepath to vocab json or directory
                containing it.
            max_token_sequence_length (int): initial value for keeping track
                of longest sequence. Defaults to 0.
            canonical (bool): performs canonicalization of SMILES (one
                original string for one molecule), if True, then other
                transformations (augment etc, see below) do not apply
            augment (bool): perform SMILES augmentation. Defaults to False.
            kekulize (bool): kekulizes SMILES (implicit aromaticity only).
                Defaults to False.
            all_bonds_explicit (bool): Makes all bonds explicit. Defaults to
                False, only applies if kekulize = True.
            all_hs_explicit (bool): Makes all hydrogens explicit. Defaults to
                False, only applies if kekulize = True.
            randomize (bool): perform a true randomization of SMILES tokens.
                Defaults to False.
            remove_bonddir (bool): Remove directional info of bonds.
                Defaults to False.
            remove_chirality (bool): Remove chirality information.
                Defaults to False.
            selfies (bool): Whether selfies is used instead of smiles, defaults
                to False.
            sanitize (bool): Sanitize SMILES. Defaults to True.
            add_start_and_stop (bool): add start and stop token indexes.
                Defaults to False.
            padding (bool): pad sequences from the left to matching length.
                Defaults to False.
            padding_length (int): common length of all token sequences,
                applies only if padding is True. See `set_max_padding` to set
                it to longest token sequence the smiles language encountered.
                Defaults to None.
            device (torch.device): device where the tensors are stored.
                Defaults to gpu, if available.

        NOTE:
            See `set_smiles_transforms` and `set_encoding_transforms` to change
            the transforms temporarily and reset with
            `reset_initial_transforms`. Assignment of class attributes
            in the parameter list will trigger such a reset.
        """
        super().__init__(
            name=name,
            smiles_tokenizer=smiles_tokenizer,
            tokenizer_name=tokenizer_name,
            vocab_file=vocab_file,
            max_token_sequence_length=max_token_sequence_length,
        )
        # smiles transforms
        self.canonical = canonical
        self.augment = augment
        self.kekulize = kekulize
        self.all_bonds_explicit = all_bonds_explicit
        self.all_hs_explicit = all_hs_explicit
        self.remove_bonddir = remove_bonddir
        self.remove_chirality = remove_chirality
        self.selfies = selfies
        self.sanitize = sanitize
        # encoding transforms
        self.randomize = randomize
        self.add_start_and_stop = add_start_and_stop
        self.padding = padding
        self.padding_length = padding_length
        self.device = device

        self._init_attributes = [  # additions to init_kwargs for pretrained
            'canonical',
            'augment',
            'kekulize',
            'all_bonds_explicit',
            'all_hs_explicit',
            'remove_bonddir',
            'remove_chirality',
            'selfies',
            'sanitize',
            'randomize',
            'add_start_and_stop',
            'padding',
            'padding_length',
        ]
        # update save/load pretrained kwargs
        for keyword in self._init_attributes:
            self.init_kwargs[keyword] = getattr(self, keyword)

        self.reset_initial_transforms()

        self._attributes_to_trigger_reset = [
            *self._init_attributes,
            'device',
            'start_index',
            'stop_index',
        ]  # could be updated in inheritance

        # only now 'activate' setter that resets the transforms and warns on
        # truncating padding_length
        self._initialized = True

    def __setattr__(self, name, value):
        """Also updates the transforms if the set attribute affects them."""
        super().__setattr__(name, value)
        if self.__dict__.get('_initialized'):
            if name in self._attributes_to_trigger_reset:
                self.reset_initial_transforms()
                if name in self._init_attributes:
                    self.init_kwargs[name] = value
            if name == 'padding_length' and self.padding:
                if self.max_token_sequence_length > value:
                    logger.warning(
                        'The language has seen sequences of length '
                        f'{self.max_token_sequence_length} that will be '
                        'truncated by given padding length of '
                        f'{value}. Consider `set_max_padding`.'
                    )

    @staticmethod
    def __get_total_number_of_tokens_with_start_stop_fn(x):
        return len(x) + 2

    def _set_token_len_fn(self, add_start_and_stop):
        """
        Defines a Callable that given a sequence of naive tokens, i.e. before
        applying the encoding transforms, computes the number of
        implicit tokens after transforms (implicit because it's the
        number of token indexes, not actual tokens).
        """
        if add_start_and_stop:
            self._get_total_number_of_tokens_fn = (
                self.__get_total_number_of_tokens_with_start_stop_fn
            )
        else:
            self._get_total_number_of_tokens_fn = len

    def set_max_padding(self):
        """
        Set padding_length that does not truncate any sequence. Requires
        updated max_token_sequence_length.

        Raises:
            UnknownMaxLengthError: When max_token_sequence_length is 0 because
                no SMILES were added to the language.
        """
        if self.max_token_sequence_length == 0:
            raise UnknownMaxLengthError(
                'No check possible for naive SMILESTokenizer. Instance needs '
                'a pass over the data, setting max_token_sequence_length. '
                'See for example `add_smis`, `add_dataset` or `add_smiles` '
                'methods.'
            )

        # also triggers reset of transforms
        self.padding_length = self.max_token_sequence_length

    def reset_initial_transforms(self):
        """Reset smiles and token indexes transforms as on initialization."""
        self.transform_smiles = compose_smiles_transforms(
            self.canonical,
            self.augment,
            self.kekulize,
            self.all_bonds_explicit,
            self.all_hs_explicit,
            self.remove_bonddir,
            self.remove_chirality,
            self.selfies,
            self.sanitize,
        )
        self.transform_encoding = compose_encoding_transforms(
            self.randomize,
            self.add_start_and_stop,
            self.start_index,
            self.stop_index,
            self.padding,
            self.padding_length,
            self.padding_index,
            self.device,
        )
        self._set_token_len_fn(self.add_start_and_stop)

    def set_smiles_transforms(
        self,
        canonical=None,
        augment=None,
        kekulize=None,
        all_bonds_explicit=None,
        all_hs_explicit=None,
        remove_bonddir=None,
        remove_chirality=None,
        selfies=None,
        sanitize=None,
    ):
        """Helper function to reversibly change steps of the transforms."""
        self.transform_smiles = compose_smiles_transforms(
            canonical=canonical if canonical is not None else self.canonical,
            augment=augment if augment is not None else self.augment,
            kekulize=kekulize if kekulize is not None else self.kekulize,
            all_bonds_explicit=all_bonds_explicit
            if all_bonds_explicit is not None
            else self.all_bonds_explicit,
            all_hs_explicit=all_hs_explicit
            if all_hs_explicit is not None
            else self.all_hs_explicit,
            remove_bonddir=remove_bonddir
            if remove_bonddir is not None
            else self.remove_bonddir,
            remove_chirality=remove_chirality
            if remove_chirality is not None
            else self.remove_chirality,
            selfies=selfies if selfies is not None else self.selfies,
            sanitize=sanitize if sanitize is not None else self.sanitize,
        )

    def set_encoding_transforms(
        self,
        randomize=None,
        add_start_and_stop=None,
        padding=None,
        padding_length=None,
        device=None,
    ):
        """Helper function to reversibly change steps of the transforms."""
        self.transform_encoding = compose_encoding_transforms(
            randomize=randomize if randomize is not None else self.randomize,
            add_start_and_stop=add_start_and_stop
            if add_start_and_stop is not None
            else self.add_start_and_stop,
            start_index=self.start_index,
            stop_index=self.stop_index,
            padding=padding if padding is not None else self.padding,
            padding_length=padding_length
            if padding_length is not None
            else self.padding_length,
            padding_index=self.padding_index,
            device=device if device is not None else self.device,
        )
        if add_start_and_stop is not None:
            self._set_token_len_fn(add_start_and_stop)



"""GeneExpressionDataset module."""
import torch

from ..types import GeneList, Optional
from ._table_dataset import _TableEagerDataset, _TableLazyDataset
from .base_dataset import DatasetDelegator

TABLE_DATASET_IMPLEMENTATIONS = {'eager': _TableEagerDataset, 'lazy': _TableLazyDataset}


class GeneExpressionDataset(DatasetDelegator):
    """
    Gene expression dataset implementation.
    """

    def __init__(
        self,
        *gene_expression_filepaths: str,
        gene_list: GeneList = None,
        standardize: bool = True,
        min_max: bool = False,
        processing_parameters: dict = {},
        impute: Optional[float] = 0.0,
        dtype: torch.dtype = torch.float,
        device: torch.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu'
        ),
        backend: str = 'eager',
        chunk_size: int = 10000,
        **kwargs,
    ) -> None:
        """
        Initialize a gene expression dataset.

        Args:
            gene_expression_filepaths (Files): paths to .csv files.
                Currently, the only supported format is .csv, with gene
                profiles on rows and gene names as columns.
            gene_list (GeneList): a list of genes. Defaults to None.
            standardize (bool): perform data standardization. Defaults to True.
            min_max (bool): perform min-max scaling. Defaults to False.
            processing_parameters (dict): processing parameters.
                Keys can be 'min', 'max' or 'mean', 'std'
                respectively. Values must be readable by `np.array`, and the
                required order and subset of features has to match that
                determined by the dataset setup (see `self.gene_list` after
                initialization). Defaults to {}.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to 0.0.
            dtype (torch.dtype): data type. Defaults to torch.float.
            device (torch.device): device where the tensors are stored.
                Defaults to gpu, if available.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
            chunk_size (int): size of the chunks in case of lazy reading, is
                ignored with 'eager' backend. Defaults to 10000.
            kwargs (dict): additional parameters for pd.read_csv.
        """
        if not (backend in TABLE_DATASET_IMPLEMENTATIONS):
            raise RuntimeError(
                'backend={} not supported! '.format(backend)
                + 'Select one in [{}]'.format(
                    ','.join(TABLE_DATASET_IMPLEMENTATIONS.keys())
                )
            )
        self.dataset = TABLE_DATASET_IMPLEMENTATIONS[backend](
            filepaths=gene_expression_filepaths,
            feature_list=gene_list,
            standardize=standardize,
            min_max=min_max,
            processing_parameters=processing_parameters,
            impute=impute,
            dtype=dtype,
            device=device,
            chunk_size=chunk_size,
            **kwargs,
        )
        # if it was not passed, gene_list is common subset in files.
        self.gene_list = self.dataset.feature_list
        self.number_of_features = len(self.gene_list)
        DatasetDelegator.__init__(self)  # delegate to self.dataset
